from __future__ import annotations

import dataclasses
import importlib
import importlib.machinery
import os
import re
import textwrap
from collections import OrderedDict as odict
from typing import FrozenSet, Iterable, List, Optional, Union

from contextlog import get_logger

from annet import patching, tracing
from annet.annlib.rbparser.acl import compile_acl_text
from annet.cli_args import GenSelectOptions, ShowGeneratorsOptions
from annet.lib import get_context
from annet.storage import Device
from annet.tracing import tracing_connector
from annet.types import (
    GeneratorEntireResult,
    GeneratorJSONFragmentResult,
    GeneratorPartialResult,
    GeneratorPartialRunArgs,
    GeneratorResult,
)
from annet.vendors import registry_connector, tabparser

from .base import BaseGenerator
from .base import ParamsList as ParamsList
from .base import TextGenerator as TextGenerator
from .entire import Entire
from .exceptions import GeneratorError, NotSupportedDevice
from .jsonfragment import JSONFragment
from .partial import PartialGenerator
from .perf import GeneratorPerfMesurer
from .ref import RefGenerator
from .result import RunGeneratorResult


# =====
DISABLED_TAG = "disable"


# =====
def get_list(args: ShowGeneratorsOptions):
    if args.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = args.generators_context
    return {
        cls.__class__.__name__: {
            "type": cls.TYPE,
            "tags": set(cls.TAGS),
            "description": get_description(cls.__class__),
        }
        for cls in _get_generators(get_context()["generators"], None)
    }


def get_description(gen_cls) -> str:
    return textwrap.dedent(" ".join([
        (gen_cls.__doc__ or ""),
        ("Disabled. Use '-g %s' to enable" % gen_cls.__name__ if DISABLED_TAG in gen_cls.TAGS else "")
    ])).strip()


def validate_genselect(gens: GenSelectOptions, all_classes):
    logger = get_logger()
    unknown_err = "Unknown generator alias %s"
    all_aliases = {
        alias
        for cls in all_classes
        for alias in cls.get_aliases()
    }
    for gen_set in (gens.allowed_gens, gens.force_enabled):
        for alias in set(gen_set or ()) - all_aliases:
            logger.error(unknown_err, alias)
            raise Exception(unknown_err % alias)


@dataclasses.dataclass
class Generators:
    """Collection of various types of generators."""

    partial: List[PartialGenerator] = dataclasses.field(default_factory=list)
    entire: List[Entire] = dataclasses.field(default_factory=list)
    ref: List[RefGenerator] = dataclasses.field(default_factory=list)
    json_fragment: List[JSONFragment] = dataclasses.field(default_factory=list)


def build_generators(storage, gens: GenSelectOptions, device: Optional[Device] = None) -> Generators:
    """Return generators that meet the gens filter conditions."""
    if gens.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = gens.generators_context
    all_generators = _get_generators(get_context()["generators"], storage, device)
    ref_generators = _get_ref_generators(get_context()["generators"], storage, device)
    validate_genselect(gens, all_generators)
    classes = list(select_generators(gens, all_generators))
    partial = [obj for obj in classes if obj.TYPE == "PARTIAL"]
    entire = [obj for obj in classes if obj.TYPE == "ENTIRE"]
    entire = list(sorted(entire, key=lambda x: x.prio, reverse=True))
    json_fragment = [obj for obj in classes if obj.TYPE == "JSON_FRAGMENT"]
    return Generators(
        partial=partial,
        entire=entire,
        json_fragment=json_fragment,
        ref=ref_generators,
    )


@tracing.function
def run_partial_initial(device):
    from .common.initial import InitialConfig

    tracing_connector.get().set_device_attributes(tracing_connector.get().get_current_span(), device)

    run_args = GeneratorPartialRunArgs(device)
    return run_partial_generators([InitialConfig(storage=device.storage, do_run=True)], [], run_args)


@tracing.function
def run_partial_generators(
    gens: List["PartialGenerator"],
    ref_gens: List["RefGenerator"],
    run_args: GeneratorPartialRunArgs,
):
    logger = get_logger(host=run_args.device.hostname)
    tracing_connector.get().set_device_attributes(tracing_connector.get().get_current_span(), run_args.device)

    ret = RunGeneratorResult()
    if run_args.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = run_args.generators_context

    for gen in ref_gens:
        ret.ref_matcher.add(gen.ref(run_args.device), gen)

    logger.debug("Generating selected PARTIALs ...")

    for gen in gens:
        try:
            result = _run_partial_generator(gen, run_args)
        except NotSupportedDevice as exc:
            logger.info("generator %s raised unsupported error: %r", gen, exc)
            continue

        if not result:
            continue

        ref_match = ret.ref_matcher.match(result.config)
        for ref_gen, groups in ref_match:
            gens.append(ref_gen.with_groups(groups))
            ret.ref_track.add(gen.__class__, ref_gen.__class__)

        ret.ref_track.config(gen.__class__, result.config)
        ret.add_partial(result)

    return ret


@tracing.function(name="run_partial_generator")
def _run_partial_generator(gen: "PartialGenerator", run_args: GeneratorPartialRunArgs) -> GeneratorPartialResult | None:
    logger = get_logger(generator=_make_generator_ctx(gen))
    device = run_args.device
    output = ""
    config = odict()
    safe_config = odict()

    if not gen.supports_device(device):
        logger.info("generator %s is not supported for device %s", gen, device.hostname)
        return None

    span = tracing_connector.get().get_current_span()
    if span:
        tracing_connector.get().set_device_attributes(span, run_args.device)
        tracing_connector.get().set_dimensions_attributes(span, gen, run_args.device)
        span.set_attributes({
            "use_acl": run_args.use_acl,
            "use_acl_safe": run_args.use_acl_safe,
            "generators_context": str(run_args.generators_context),
        })

    with GeneratorPerfMesurer(gen, run_args=run_args) as pm:
        if not run_args.no_new:
            if gen.get_user_runner(device):
                logger.info("Generating PARTIAL ...")
            try:
                output = gen(device, run_args.annotate)
            except NotSupportedDevice:
                # это исключение нужно передать выше в оригинальном виде
                raise
            except Exception as err:
                filename, lineno = gen.get_running_line()
                logger.error("Generator error in file '%s:%i'", filename, lineno)
                raise GeneratorError(f"{gen} on {device}") from err

            fmtr = registry_connector.get().match(device.hw).make_formatter()

            try:
                config = tabparser.parse_to_tree(text=output, splitter=fmtr.split)
            except tabparser.ParserError as err:
                logger.exception("Parser error")
                raise GeneratorError from err

    acl = gen.acl(device) or ""
    rules = compile_acl_text(textwrap.dedent(acl), device.hw.vendor)
    acl_safe = gen.acl_safe(device) or ""
    safe_rules = compile_acl_text(textwrap.dedent(acl_safe), device.hw.vendor)

    if run_args.use_acl:
        try:
            with tracing_connector.get().start_as_current_span(
                "apply_acl", tracer_name=__name__, min_duration="0.01"
            ) as acl_span:
                tracing_connector.get().set_device_attributes(acl_span, run_args.device)
                config = patching.apply_acl(
                    config=config,
                    rules=rules,
                    fatal_acl=True,
                    with_annotations=run_args.annotate,
                )
            if run_args.use_acl_safe:
                with tracing_connector.get().start_as_current_span(
                    "apply_acl_safe",
                    tracer_name=__name__,
                    min_duration="0.01"
                ) as acl_safe_span:
                    tracing_connector.get().set_device_attributes(acl_safe_span, run_args.device)
                    safe_config = patching.apply_acl(
                        config=config,
                        rules=safe_rules,
                        fatal_acl=False,
                        with_annotations=run_args.annotate,
                    )
        except patching.AclError as err:
            logger.error("ACL error: generator is not allowed to yield this command: %s", err)
            raise GeneratorError from err
        except NotImplementedError as err:
            logger.error(str(err))
            raise GeneratorError from err

    return GeneratorPartialResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        output=output,
        acl=acl,
        acl_rules=rules,
        acl_safe=acl_safe,
        acl_safe_rules=safe_rules,
        config=config,
        safe_config=safe_config,
        perf=pm.last_result,
    )


@tracing.function
def check_entire_generators_required_packages(gens, device_packages: FrozenSet[str]) -> List[str]:
    errors: List[str] = []
    for gen in gens:
        if not gen.REQUIRED_PACKAGES.issubset(device_packages):
            missing = gen.REQUIRED_PACKAGES - device_packages
            missing_str = ", ".join("`{}'".format(pkg) for pkg in sorted(missing))
            if len(missing) == 1:
                errors.append("missing package {} required for {}".format(missing_str, gen))
            else:
                errors.append("missing packages {} required for {}".format(missing_str, gen))
    return errors


@tracing.function
def run_file_generators(
    gens: Iterable[Union["JSONFragment", "Entire"]],
    device: "Device",
) -> RunGeneratorResult:
    """Run generators that generate files or file parts."""
    ret = RunGeneratorResult()
    logger = get_logger(host=device.hostname)
    logger.debug("Generating selected ENTIREs and JSON_FRAGMENTs ...")
    for gen in gens:
        if gen.__class__.TYPE == "ENTIRE":
            run_generator_fn = _run_entire_generator
            add_result_fn = ret.add_entire
        elif gen.__class__.TYPE == "JSON_FRAGMENT":
            run_generator_fn = _run_json_fragment_generator
            add_result_fn = ret.add_json_fragment
        else:
            raise RuntimeError(f"Unknown generator class type: cls={gen.__class__} TYPE={gen.__class__.TYPE}")
        try:
            result = run_generator_fn(gen, device)
        except NotSupportedDevice as exc:
            logger.info("generator %s raised unsupported error: %r", gen, exc)
            continue
        if result:
            add_result_fn(result)

    return ret


@tracing.function(min_duration="0.5")
def _run_entire_generator(gen: "Entire", device: "Device") -> Optional[GeneratorResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))
    span = tracing_connector.get().get_current_span()
    if span:
        tracing_connector.get().set_device_attributes(span, device)
        tracing_connector.get().set_dimensions_attributes(span, gen, device)

    with GeneratorPerfMesurer(gen, trace_min_duration="0.5") as pm:
        if not gen.supports_device(device):
            logger.debug("generator %s is not supported for device %s", gen, device.hostname)
            return

        path = gen.path(device)
        if not path:
            raise RuntimeError("entire generator should return non-empty path")

        logger.info("Generating ENTIRE ...")

        gen_output = gen(device)
        gen_reload = gen.get_reload_cmds(device)
        gen_is_safe = gen.is_safe(device)
        gen_prio = gen.prio

    return GeneratorEntireResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        path=path,
        output=gen_output,
        reload=gen_reload,
        prio=gen_prio,
        perf=pm.last_result,
        is_safe=gen_is_safe,
    )


def _make_generator_ctx(gen):
    return "%s.[%s]" % (gen.__module__, gen.__class__.__name__)


def _run_json_fragment_generator(
    gen: "JSONFragment",
    device: "Device",
) -> Optional[GeneratorResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))

    with GeneratorPerfMesurer(gen) as pm:
        if not gen.supports_device(device):
            logger.info("generator %s is not supported for device %s", gen, device.hostname)
            return

        path = gen.path(device)
        if not path:
            raise RuntimeError("json fragment generator should return non-empty path")

        acl_item_or_list_of_items = gen.acl(device)
        safe_acl_item_or_list_of_items = gen.acl_safe(device)
        if not acl_item_or_list_of_items:
            raise RuntimeError("json fragment generator should return non-empty acl")
        if isinstance(acl_item_or_list_of_items, list):
            acl = acl_item_or_list_of_items
        else:
            acl = [acl_item_or_list_of_items]
        if isinstance(safe_acl_item_or_list_of_items, list):
            acl_safe = safe_acl_item_or_list_of_items
        else:
            acl_safe = [safe_acl_item_or_list_of_items]

        logger.info("Generating JSON_FRAGMENT ...")

        config = gen(device)
        reload_cmds = gen.get_reload_cmds(device)

    return GeneratorJSONFragmentResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        path=path,
        acl=acl,
        acl_safe=acl_safe,
        config=config,
        reload=reload_cmds,
        perf=pm.last_result,
        reload_prio=gen.reload_prio,
    )


def _get_generators(module_paths: Union[List[str], dict], storage, device=None):
    if isinstance(module_paths, dict):
        if device is None:
            module_paths = module_paths.get("default", [])
        else:
            modules = []
            seen = set()
            matched_property = None
            for prop, prop_modules in module_paths.get("per_device_property", {}).items():
                if getattr(device, prop, False) is True:
                    if matched_property is not None:
                        raise RuntimeError(
                            f"Device {device.hostname} is matched by more than one "
                            f"per_device_property: {matched_property} and {prop}"
                        )
                    matched_property = prop
                    for module in prop_modules:
                        if module not in seen:
                            modules.append(module)
                            seen.add(module)
            module_paths = modules or module_paths.get("default", [])
    res_generators = []
    for module_path in module_paths:
        module = _load_gen_module(module_path)
        if hasattr(module, "get_generators"):
            generators: List[BaseGenerator] = module.get_generators(storage)
            res_generators += generators
    return res_generators


def _load_gen_module(module_path: str):
    try:
        module = importlib.import_module(module_path)
    except ModuleNotFoundError as e:
        try:  # maybe it's a path to module
            module_abs_path = os.path.abspath(module_path)
            module = importlib.machinery.SourceFileLoader(re.sub(r"[./]", "_", module_abs_path).strip("_"),
                                                          module_abs_path).load_module()
        except ModuleNotFoundError:
            raise e
    return module


def _get_ref_generators(module_paths: List[str], storage, device):
    if isinstance(module_paths, dict):
        module_paths = module_paths.get("default")
    res_generators = []
    for module_path in module_paths:
        module = _load_gen_module(module_path)
        if hasattr(module, "get_ref_generators"):
            res_generators += module.get_ref_generators(storage)
    return res_generators


def select_generators(gens: GenSelectOptions, classes: Iterable[BaseGenerator]):
    def contains(obj, where):
        if where:
            return obj.get_aliases().intersection(where)
        return False

    def has(cls, what):
        return what in cls.TAGS

    flts = [lambda c: not isinstance(c, RefGenerator)]
    if gens.allowed_gens:
        flts.append(lambda c: contains(c, gens.allowed_gens))
    elif gens.force_enabled:
        flts.append(lambda c: not has(c, DISABLED_TAG) or contains(c, gens.force_enabled))
    elif not gens.ignore_disabled:
        flts.append(lambda c: not has(c, DISABLED_TAG))

    if gens.excluded_gens:
        flts.append(lambda c: not contains(c, gens.excluded_gens))

    return filter(lambda x: all(f(x) for f in flts), classes)



import abc
import bisect
import contextlib
import functools
import ipaddress
import itertools
import math
import os
import pathlib
import re
import socket
import sys
import textwrap
import typing
from collections import OrderedDict as odict
from collections.abc import Iterable
from functools import lru_cache
from typing import List, NamedTuple, Optional, Tuple, Union

import contextlog
import mako.template

_logger = contextlog.get_logger()


# =====
class HuaweiNumBlock(metaclass=abc.ABCMeta):
    def __init__(self, generator, block_name, start=5, step=5):
        self._gen = generator
        self._block_name = block_name
        self._rule_num = start
        self._rule_step = step
        self._block = None

    def _current(self):
        (current, self._rule_num) = (self._rule_num, self._rule_num + self._rule_step)
        return current

    def __enter__(self):
        self._block = self._gen.block(self._block_name)
        self._block.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, trace):
        self._block.__exit__(exc_type, exc_val, trace)
        self._block = None

    def __call__(self, rule_str, min_number=None):
        if min_number is not None:
            self._rule_num = max(self._rule_num, int(min_number))
        return self.format_rule_string(rule_str)

    @abc.abstractmethod
    def format_rule_string(self, rule_str):
        pass


def huawei_expand_vlandb(row):
    expanded = set()
    row_parts = row.split()
    for (index, part) in enumerate(row_parts):
        if part == "to":
            left = int(row_parts[index - 1])
            right = int(row_parts[index + 1])
            expanded = expanded.union(range(left + 1, right))
        else:
            expanded.add(int(part))
    return expanded


def cisco_expand_vlandb(row):
    expanded = set()
    for part in row.split(","):
        range_parts = [int(p.strip()) for p in part.strip().split("-")]
        assert len(range_parts) in (1, 2)
        if len(range_parts) > 1:
            left = range_parts[0]
            right = range_parts[1]
            expanded.update(range(left, right + 1))
        else:
            expanded.add(int(range_parts[0]))
    return expanded


def huawei_collapse_vlandb(vlans, chunk_len=0):
    return collapse_vlandb(vlans, " to ", chunk_len=chunk_len)


def cisco_collapse_vlandb(vlans, tiny_ranges=True):
    return collapse_vlandb(vlans, "-", tiny_ranges)


def juniper_fmt_prefix_lists_acl(prefix_list_names: typing.Iterable[str]) -> str:
    """Форматирует строку для acl из prefix-list-имён.

    Производит следующие операции:

    - вырезает дубликаты
    - сортирует
    - добавляет quoted-вариант имени (NOCDEV-5321)
    - соединяет всё в строку
    """
    sorted_unique_names = sorted(frozenset(prefix_list_names))
    quoted_and_unquoted_names = []
    for prefix_list_name in sorted_unique_names:
        quoted_and_unquoted_names.append(prefix_list_name)
        quoted_and_unquoted_names.append('"{}"'.format(prefix_list_name))
    joined_prefix_lists = "|".join(quoted_and_unquoted_names)
    return joined_prefix_lists


# tiny_ranges - если выставлен в False, то два рядом стоящих влана не конвертятся в
# рендж (поведение cisco catalyst)
def collapse_vlandb(vlans, range_sep, tiny_ranges=True, chunk_len=0):
    assert len(vlans) != 0
    vlans = sorted(set(vlans))
    res = []
    row = [vlans[0], vlans[0]]
    for vlan in vlans[1::]:
        if row[1] == (vlan - 1):
            row[1] = vlan
        elif not tiny_ranges and row[1] - row[0] == 1:
            res.append([row[0], row[0]])
            res.append([row[1], row[1]])
            row = [vlan, vlan]
        else:
            res.append(row)
            row = [vlan, vlan]
    res.append(row)
    res = list(map(lambda x: x[0] != x[1] and "%s%s%s" % (x[0], range_sep, x[1]) or str(x[0]), res))

    # Устройства бьют списки вланов на чанки при добавлении в конфиг
    if chunk_len:
        chunks = [res[i:i + chunk_len] for i in range(0, len(res), chunk_len)]
        return chunks

    return res


def huawei_iface_ranges(iface_names):
    ret = []
    grouped = odict()
    for iface in iface_names:
        match = re.match(r"(.*[:/])(\d+)$", iface)
        if not match:
            ret.append(iface)
        else:
            prefix, index = match.groups()
            grouped.setdefault(prefix, []).append((int(index), iface))

    for ifaces in grouped.values():
        start = 0
        while start < len(ifaces):
            end = start
            while end + 1 < len(ifaces) and ifaces[end + 1][0] == ifaces[end][0] + 1:
                end += 1
            if start == end:
                ret.append(ifaces[start][1])
            else:
                ret.append((ifaces[start][1], ifaces[end][1]))
            start = end + 1

    return ret


def juniper_port_split(iface_name):
    ret = dict(zip(["speed", "fpc", "pic", "port"], re.split(r"-+|/+|:", iface_name)))
    return ret


def make_ip4_mask(prefix_len, inverse=False):
    """
    Returns dotted-quad string representing IPv4 mask of given length

    E.g., for prefix_len=25 it returns "255.255.255.128"
    """
    if not isinstance(prefix_len, int) or prefix_len < 0 or prefix_len > 32:
        raise ValueError("invalid prefix_len %r" % prefix_len)
    bin_mask = (0xffffffff & (0xffffffff << (32 - prefix_len))).to_bytes(4, byteorder="big")
    if inverse:
        bin_mask = bytes(byte ^ 255 for byte in bin_mask)
    return socket.inet_ntop(socket.AF_INET, bin_mask)


def merge_dicts(*args):
    if len(args) == 0:
        return odict()
    if len(args) == 1:
        return args[0]
    if all(map(lambda x: x[0] == x[1], zip(args[:-1], args[1:]))):
        return args[0]

    merged = odict()
    for dictionary in args:
        assert isinstance(dictionary, (dict, odict))
        for (key, value) in dictionary.items():
            if isinstance(value, (dict, odict)):
                merged[key] = merge_dicts(*[x[key] for x in args if key in x])
            elif key not in merged:
                merged[key] = value
            elif isinstance(merged[key], (list, tuple)):
                merged[key] = merged[key].__class__(itertools.chain(merged[key], value))
            else:
                merged[key] = value
    return merged


def meta_decor(*method_list):
    def wrapper(*args, **kwds):
        for method in method_list:
            res = method(*args, **kwds)
        return res

    return wrapper


def first(iterable, default=None):
    return next(iter(iterable), default)


def flatten(iterable):
    for x in iterable:
        if not isinstance(x, (str, bytes)) and isinstance(x, Iterable):
            yield from flatten(x)
        else:
            yield x


def uniq(*iterables):
    seen = set()
    for iterable in iterables:
        for item in iterable:
            if item not in seen:
                seen.add(item)
                yield item


# Хелпер чтобы при итерировании по вложенным структурам
# правильно подбирать значения и флаги (local_as, update_source, multipath, etc)
# которые должны наследоваться от верхних блоков к нижним
# XXX Покрыть тестами
class ContextOrderedDict:
    def __init__(self, iterable):
        self.dict = odict(iterable)

    @contextlib.contextmanager
    def update(self, peers_sub_dict):
        values_backup = self.dict.copy()
        for key in set(self.dict).intersection(peers_sub_dict.keys()):
            if not isinstance(peers_sub_dict[key], dict):
                self.dict[key] = peers_sub_dict[key]
        yield
        self.dict = values_backup
        del values_backup


# {{{ http://code.activestate.com/recipes/511478/ (r1)
def percentile(list_numbers, percent, key=lambda x: x):
    """
    Find the percentile of a list of values.

    @parameter N - is a list of values.
    @parameter percent - a float value from 0.0 to 1.0.
    @parameter key - optional key function to compute value from each element of N.

    @return - the percentile of the values
    """
    if not list_numbers:
        return None
    list_numbers = sorted(list_numbers, key=key)
    kkk = (len(list_numbers) - 1) * percent
    floo = math.floor(kkk)
    cei = math.ceil(kkk)
    if floo == cei:
        return key(list_numbers[int(kkk)])
    d0 = key(list_numbers[int(floo)]) * (cei - kkk)
    d1 = key(list_numbers[int(cei)]) * (kkk - floo)
    return d0 + d1


def mako_render(template, dedent=False, **kwargs):
    @lru_cache(None)
    def _compile_mako(template, dedent):
        if dedent:
            template = textwrap.dedent(template).strip()
        return mako.template.Template(template)

    ret = _compile_mako(template, dedent).render(**kwargs)
    if dedent:
        ret = ret.strip()
    return ret


# =====
def find_exc_in_stack(
    container_exc: Exception,
    target_exc_type: Union[type, Tuple[type, ...]]
) -> Optional[BaseException]:
    curr_err: Optional[BaseException] = container_exc
    while curr_err is not None:
        if isinstance(curr_err, target_exc_type):
            return curr_err
        curr_err = curr_err.__context__
    return None


def find_modules(base_dir):
    """
    Рекурсивно ищем в base_dir файлы python модулей
    """
    for entry in os.scandir(base_dir):
        if entry.name[:1] in ["_", "."]:
            continue

        if entry.is_dir():
            child_dir = os.path.join(base_dir, entry.name)
            yield from (".".join((entry.name, child_name)) for child_name in find_modules(child_dir))
        elif entry.is_file() and entry.name.endswith(".py"):
            yield entry.name[:-3]


def catch_ctrl_c(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except KeyboardInterrupt:
            _logger.error("Got Ctrl+C, exiting")
            sys.exit(127)

    return wrapper


class LMSegment(NamedTuple):
    """
    LMSegment хранит подсеть в виде целочисленных пар (int(network_addr), int(broadcast_addr)):
    0.0.0.0/0:  (0b0000000000000000000000000, 0b1111111111111111111111111)
    0.0.0.0/16: (0b0000000000000000000000000, 0b0000000000001111111111111)
    1.0.0.0/24: (0b1000000000000000000000000, 0b1000000000000000011111111)

    Для двух x=(A, B), y=(C, D) таких пар (A <= B, C <= D по определению):

    Отношение включения: x contains y <-> A <= C && D <= B:
    1.0.0.0/24 contains 1.0.0.1/32
    0.0.0.0/0 contains 0.0.0.0/16

    Отношение порядка: x < y <-> A < C || (A == C && B > D), или проще (A, -B) < (C, -D):
    1.0.0.0/24 < 1.0.0.1/32
    0.0.0.0/0  < 0.0.0.0/16
    """

    begin: int
    end: int

    def contains(self, other: "LMSegment") -> bool:
        return self.begin <= other.begin and other.end <= self.end

    @classmethod
    def from_net(cls, net: Union[ipaddress.IPv4Network, ipaddress.IPv6Network]):
        return cls(
            begin=int(net.network_address),
            end=int(net.network_address) | int(net.hostmask),
        )

    def _cmp_identity(self) -> Tuple[int, int]:
        return (self.begin, -self.end)

    def __eq__(self, other: "LMSegment"):
        return self._cmp_identity() == other._cmp_identity()

    def __lt__(self, other: "LMSegment"):
        return self._cmp_identity() < other._cmp_identity()

    def __le__(self, other: "LMSegment"):
        return self._cmp_identity() <= other._cmp_identity()

    def __ge__(self, other: "LMSegment"):
        return self._cmp_identity() >= other._cmp_identity()

    def __gt__(self, other: "LMSegment"):
        return self._cmp_identity() > other._cmp_identity()


class LMSegmentList:
    """Упорядоченный список подсетей
    """

    def __init__(self):
        self.pfxs: List[LMSegment] = []

    def add(self, pref: LMSegment) -> None:
        """Добавляем новый префикс в упорядоченный список, если он не дублируется
        """
        idx = bisect.bisect(self.pfxs, pref) - 1
        if 0 <= idx < len(self.pfxs):
            if pref == self.pfxs[idx]:
                return
        self.pfxs.insert(idx + 1, pref)

    def find(self, target: LMSegment) -> Optional[LMSegment]:
        """LPM поиск в добавленных
        """
        upper_bound = bisect.bisect(self.pfxs, target)
        for i in reversed(range(0, upper_bound)):
            if self.pfxs[i].contains(target):
                return self.pfxs[i]
            if self.pfxs[i] > target:
                break
        return None


class LMSMatcher:
    """Обертка над парой LMSegmentList над парой LMSegmentList для v4/v6
    """

    def __init__(self):
        self.v4 = LMSegmentList()
        self.v6 = LMSegmentList()

    def add(self, pref: str) -> None:
        net = ipaddress.ip_network(pref)
        self._af(net).add(LMSegment.from_net(net))

    def find(self, pref: str) -> Optional[str]:
        net = ipaddress.ip_network(pref)
        found = self._af(net).find(LMSegment.from_net(net))
        if found:
            net_addr = net.__class__(found.begin).network_address
            net_hostmask = (found.begin ^ found.end).bit_length()
            net_mask = net_addr.max_prefixlen - net_hostmask
            return str(net_addr) + "/" + str(net_mask)

    def _af(self, net):
        if net.version == 4:
            return self.v4
        elif net.version == 6:
            return self.v6
        else:
            raise ValueError(str(net))


def is_relative(path: pathlib.PurePath, *other: str) -> bool:
    """Проверяет является ли путь path относительным любого из other
    """
    for checkpath in other:
        try:
            path.relative_to(checkpath)
            return True
        except ValueError:
            pass
    return False


_ANNOTATION_SEP = "\t# "


def add_annotation(row: str, annotation: str) -> str:
    return _ANNOTATION_SEP.join((row, annotation))


def strip_annotation(row: str) -> str:
    return row.rsplit(_ANNOTATION_SEP, 1)[0]


def jun_inactive_pfx() -> str:
    return "inactive: "


def jun_is_inactive(key) -> bool:
    return key.startswith(jun_inactive_pfx())


def jun_activate(key) -> str:
    if jun_is_inactive(key):
        key = key[len(jun_inactive_pfx()):]
    return key


# TODO: Remove with python3.10
async def anext(async_iter):
    return await async_iter.__anext__()  # pylint: disable=unnecessary-dunder-call



from typing import Any

from annet.generators import BaseGenerator, Entire
from annet.mesh import MeshExecutor
from annet.rpl import RouteMap, RoutingPolicy
from annet.rpl_generators import (
    CommunityListGenerator, RoutingPolicyGenerator, AsPathFilterGenerator, CommunityList, AsPathFilter,
    RDFilterFilterGenerator, RDFilter, PrefixListFilterGenerator, IpPrefixList, CumulusPolicyGenerator,
    get_policies,
)
from annet.storage import Storage
from .items import COMMUNITIES, AS_PATH_FILTERS, RD_FILTERS, PREFIX_LISTS
from .mesh import registry
from .route_policy import routemap


class CommunityGenerator(CommunityListGenerator):
    def get_community_lists(self, device: Any) -> list[CommunityList]:
        return COMMUNITIES

    def get_routemap(self) -> RouteMap:
        return routemap

    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        return get_policies(
            routemap=routemap,
            device=device,
            mesh_executor=MeshExecutor(registry, self.storage),
        )


class PolicyGenerator(RoutingPolicyGenerator):
    def get_prefix_lists(self, device: Any) -> list[IpPrefixList]:
        return PREFIX_LISTS

    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        return get_policies(
            routemap=routemap,
            device=device,
            mesh_executor=MeshExecutor(registry, self.storage),
        )

    def get_community_lists(self, device: Any) -> list[CommunityList]:
        return COMMUNITIES

    def get_rd_filters(self, device: Any) -> list[RDFilter]:
        return RD_FILTERS


class AsPathGenerator(AsPathFilterGenerator):
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        return get_policies(
            routemap=routemap,
            device=device,
            mesh_executor=MeshExecutor(registry, self.storage),
        )

    def get_as_path_filters(self, device: Any) -> list[AsPathFilter]:
        return AS_PATH_FILTERS


class RDGenerator(RDFilterFilterGenerator):
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        return get_policies(
            routemap=routemap,
            device=device,
            mesh_executor=MeshExecutor(registry, self.storage),
        )

    def get_rd_filters(self, device: Any) -> list[RDFilter]:
        return RD_FILTERS


class PrefixListGenerator(PrefixListFilterGenerator):
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        return get_policies(
            routemap=routemap,
            device=device,
            mesh_executor=MeshExecutor(registry, self.storage),
        )

    def get_prefix_lists(self, device: Any) -> list[IpPrefixList]:
        return PREFIX_LISTS


FRR_HEADER = """\
!!! This file is auto-generated
!
frr defaults datacenter
log syslog informational
log timestamp precision 6
service integrated-vtysh-config"""


class FrrGenerator(Entire, CumulusPolicyGenerator):
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        return get_policies(
            routemap=routemap,
            device=device,
            mesh_executor=MeshExecutor(registry, self.storage),
        )

    def get_community_lists(self, device: Any) -> list[CommunityList]:
        return COMMUNITIES

    def get_prefix_lists(self, device: Any) -> list[IpPrefixList]:
        return PREFIX_LISTS

    def get_as_path_filters(self, device: Any) -> list[AsPathFilter]:
        return AS_PATH_FILTERS

    def path(self, device):
        if device.hw.PC.Mellanox or device.hw.PC.NVIDIA:
            return "/etc/frr/frr.conf"

    def run(self, device):
        yield FRR_HEADER
        yield from self.generate_cumulus_rpl(device)
        yield "line vty"


def get_generators(store: Storage) -> list[BaseGenerator]:
    return [
        AsPathGenerator(store),
        PolicyGenerator(store),
        CommunityGenerator(store),
        RDGenerator(store),
        PrefixListGenerator(store),
        FrrGenerator(store),
    ]



import collections
import functools


def get_db(prepared):
    allowed = _make_allowed_by_seq(prepared)
    return (
        _build_tree(prepared, allowed),
        functools.reduce(set.union, allowed.values()),  # All sequences
    )


def find_true_sequences(hw_model, tree):
    sequences = set()
    for (regexp, meta) in tree.items():
        if regexp.search(hw_model):
            sequences.update(meta["sequences"])
            sequences.update(find_true_sequences(hw_model, meta["children"]))
    return sequences


def _build_tree(prepared, allowed_by_seq):
    tree = {}
    for (seq, regexp) in prepared.items():
        sub = tree
        for sub_seq in _seq_subs(seq):
            regexp = prepared[sub_seq]
            if regexp not in sub:
                sub[regexp] = {
                    "sequences": allowed_by_seq[sub_seq],
                    "children": {},
                }
            sub = sub[regexp]["children"]
    return tree


def _seq_subs(seq):
    for index in range(1, len(seq) + 1):
        yield seq[:index]


def _make_allowed_by_seq(sequences):
    all_variants = collections.Counter()
    variants_by_seq = {}

    for seq in sequences:
        variants = _make_seq_variants(seq)
        all_variants.update(variants)
        variants_by_seq[seq] = variants

    return {
        seq: set(variant for variant in variants if all_variants[variant] <= 1)
        for (seq, variants) in variants_by_seq.items()
    }


def _make_seq_variants(seq):
    return set(
        seq[left:-right] + (seq[-1],)
        for left in range(len(seq))
        for right in range(1, len(seq[left:]) + 1)
    )



from abc import abstractmethod, ABC
from collections.abc import Sequence
from typing import Any

from annet.generators import PartialGenerator
from annet.rpl import RouteMap, MatchField, RoutingPolicy
from .entities import RDFilter


class RDFilterFilterGenerator(PartialGenerator, ABC):
    TAGS = ["policy", "rpl", "routing"]

    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_rd_filters(self, device: Any) -> Sequence[RDFilter]:
        raise NotImplementedError()

    def get_used_rd_filters(self, device: Any) -> Sequence[RDFilter]:
        filters = {c.name: c for c in self.get_rd_filters(device)}
        policies = self.get_policies(device)
        used_filters = set()
        for policy in policies:
            for statement in policy.statements:
                for condition in statement.match.find_all(MatchField.rd):
                    used_filters.update(condition.value)
        return [filters[name] for name in sorted(used_filters)]

    def acl_huawei(self, _):
        return r"""
        ip rd-filter
        """

    def run_huawei(self, device: Any):
        for rd_filter in self.get_used_rd_filters(device):
            for i, route_distinguisher in enumerate(rd_filter.members):
                rd_id = (i + 1) * 10 + 5
                yield "ip rd-filter", rd_filter.number, f"index {rd_id}", "permit", route_distinguisher

    def acl_iosxr(self, _):
        return r"""
        rd-set *
            ~ %global=1
        """

    def run_iosxr(self, device: Any):
        for rd_filter in self.get_used_rd_filters(device):
            with self.block("rd-set", rd_filter.name):
                for i, route_distinguisher in enumerate(rd_filter.members):
                    if i + 1 < len(rd_filter.members):
                        comma = ","
                    else:
                        comma = ""
                    yield f"{route_distinguisher}{comma}",

    def acl_juniper(self, _):
        return r"""
        policy-options             %cant_delete
            route-distinguisher
        """

    def run_juniper(self, device: Any):
        for rd_filter in self.get_used_rd_filters(device):
            with self.block("policy-options"):
                if len(rd_filter.members) == 1:
                    yield "route-distinguisher", rd_filter.name, "members", rd_filter.members[0]
                elif len(rd_filter.members) > 1:
                    joined = " ".join(rd_filter.members)
                    yield "route-distinguisher", rd_filter.name, "members", f"[ {joined} ]"



__all__ = [
    "GlobalOptions",
    "MeshSession",
    "DirectPeer",
    "IndirectPeer",
    "MeshExecutor",
    "MeshRulesRegistry",
    "Left",
    "Right",
    "Match",
    "VirtualLocal",
    "VirtualPeer",
    "PortProcessor",
    "separate_ports",
    "united_ports"
]

from .executor import MeshExecutor
from .match_args import Left, Right, Match
from .registry import (
    DirectPeer, IndirectPeer, MeshSession, GlobalOptions, MeshRulesRegistry, VirtualLocal, VirtualPeer,
)
from .port_processor import PortProcessor, united_ports, separate_ports



import contextlib
import inspect
from abc import ABC, abstractmethod
from functools import wraps
from typing import Optional, Type, Union

from annet.connectors import CachedConnector

MinDurationT = Optional[Union[str, int]]


class _Nop:
    def __getattr__(self, item):
        return _Nop()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self, *args, **kwargs):
        return _Nop()

    def __bool__(self):
        return False


class Tracing(ABC):
    enabled = False

    @abstractmethod
    def function(self, arg=None, /, **kwargs):
        pass

    @abstractmethod
    def contextmanager(self, arg=None, /, **kwargs):
        pass

    @abstractmethod
    def get_current_span(self, context_value=None):
        pass

    @abstractmethod
    def set_device_attributes(self, span, device):
        pass

    @abstractmethod
    def start_as_current_span(self, *args, **kwargs):
        pass

    @abstractmethod
    def start_as_linked_span(self, *args, **kwargs):
        pass

    @abstractmethod
    def attach_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def inject_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def extract_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def get_context(self, *args, **kwargs):
        pass

    @abstractmethod
    def force_flush(self):
        pass


class DummyTracing(Tracing):
    def function(self, arg=None, /, **kwargs):
        def decorator(func):
            return func

        return decorator if arg is None else decorator(arg)

    def contextmanager(self, arg=None, /, **kwargs):
        def decorator(cls_or_func):
            return cls_or_func

        return decorator if arg is None else decorator(arg)

    def get_current_span(self, context_value=None):
        return None

    def set_device_attributes(self, span, device):
        return None

    def start_as_current_span(self, *args, **kwargs):
        return _Nop()

    def start_as_linked_span(self, *args, **kwargs):
        return _Nop()

    def attach_context(self, *args, **kwargs):
        return

    def inject_context(self, *args, **kwargs):
        return

    def extract_context(self, *args, **kwargs):
        return

    def get_context(self, *args, **kwargs):
        return

    def force_flush(self):
        return


class _TracingConnector(CachedConnector[Tracing]):
    name = "Tracing"
    ep_name = "tracing"

    def _get_default(self) -> Type[Tracing]:
        return DummyTracing


tracing_connector = _TracingConnector()


def function(arg=None, /, **outer_kwargs):
    def decorator(func):
        cache = None

        @wraps(func)
        def wrapper(*args, **kwargs):
            nonlocal cache
            if cache is None:
                cache = tracing_connector.get().function(func, **outer_kwargs)
            return cache(*args, **kwargs)

        return wrapper

    return decorator if arg is None else decorator(arg)


def contextmanager(arg=None, /, **outer_kwargs):
    def decorator(cls_or_func):
        cache = None

        if inspect.isfunction(cls_or_func):
            @contextlib.contextmanager
            @wraps(cls_or_func)
            def wrapper(*args, **kwargs):
                nonlocal cache
                if cache is None:
                    cache = tracing_connector.get().contextmanager(cls_or_func, **outer_kwargs)
                with cache(*args, **kwargs) as val:
                    yield val

            return wrapper
        else:
            original_enter = cls_or_func.__enter__
            original_exit = cls_or_func.__exit__

            @wraps(original_enter)
            def one_shot_enter(*args, **kwargs):
                """Wrap after first call, tracing_connector was set up"""
                nonlocal cache
                if cache is None:
                    cls_or_func.__enter__ = original_enter
                    cls_or_func.__exit__ = original_exit
                    cache = tracing_connector.get().contextmanager(cls_or_func, **outer_kwargs)
                return cache.__enter__(*args, **kwargs)  # pylint: disable=unnecessary-dunder-call

            @wraps(original_exit)
            def one_shot_exit(*args, **kwargs):
                nonlocal cache
                return cache.__exit__(*args, **kwargs)

            setattr(cls_or_func, "__enter__", one_shot_enter)
            setattr(cls_or_func, "__exit__", one_shot_exit)

            return cls_or_func

    return decorator if arg is None else decorator(arg)


def class_methods(arg=None, /, **outer_kwargs):
    def decorator(cls):
        has_enter, has_exit = False, False

        for name, attr in inspect.getmembers(cls, lambda x: (
            inspect.isroutine(x)
            and not inspect.ismethoddescriptor(x)
            and not inspect.isbuiltin(x)
        )):
            if getattr(attr, "_disable_class_methods", False):
                continue
            if name == "__enter__":
                has_enter = True
            elif name == "__exit__":
                has_exit = True
            else:
                method = function(**outer_kwargs)(attr)
                if isinstance(inspect.getattr_static(cls, name), staticmethod):
                    method = staticmethod(method)
                setattr(cls, name, method)

        if all((has_enter, has_exit)):
            cls = contextmanager(**outer_kwargs)(cls)

        return cls

    return decorator if arg is None else decorator(arg)


def disable_class_methods(func):
    setattr(func, "_disable_class_methods", True)
    return func



# pylint: disable=unused-argument
import abc
import itertools
from collections import namedtuple
from typing import Dict, List, Optional, Any, OrderedDict, Tuple, Type

from contextlog import get_logger

from annet.annlib.command import Command, Question, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.annlib.rbparser.deploying import MakeMessageMatcher, Answer
from annet.cli_args import DeployOptions
from annet.connectors import Connector, get_connector_from_config
from annet.rulebook import get_rulebook, deploying
from annet.storage import Device


_DeployResultBase = namedtuple("_DeployResultBase", ("hostnames", "results", "durations", "original_states"))


class ProgressBar(abc.ABC):
    @abc.abstractmethod
    def set_content(self, tile_name: str, content: str):
        ...

    @abc.abstractmethod
    def add_content(self, tile_name: str, content: str):
        ...

    @abc.abstractmethod
    def reset_content(self, tile_name: str):
        ...

    @abc.abstractmethod
    def set_progress(self,
                     tile_name: str,
                     iteration: int,
                     total: int,
                     prefix: str = "",
                     suffix: str = "",
                     fill: str = "",
                     error: bool = False,
                     ):
        ...

    @abc.abstractmethod
    def set_exception(self, tile_name: str, cmd_exc: str, last_cmd: str, progress_max: int, content: str = "") -> None:
        ...


class DeployResult(_DeployResultBase):  # noqa: E302
    def add_results(self, results: dict[str, Exception]) -> None:
        for hostname, excs in results.items():
            self.hostnames.append(hostname)
            self.results[hostname] = excs
            self.durations[hostname] = 0.0
            self.original_states[hostname] = None


class _FetcherConnector(Connector["Fetcher"]):
    name = "Fetcher"
    ep_name = "deploy_fetcher"
    ep_by_group_only = "annet.connectors.fetcher"

    def _get_default(self) -> Type["Fetcher"]:
        # if entry points are broken, try to use direct import
        import annet.adapters.fetchers.stub.fetcher as stub_fetcher
        return stub_fetcher.StubFetcher


class _DriverConnector(Connector["DeployDriver"]):
    name = "DeployDriver"
    ep_name = "deploy_driver"
    ep_by_group_only = "annet.connectors.deployer"

    def _get_default(self) -> Type["DeployDriver"]:
        # if entry points are broken, try to use direct import
        import annet.adapters.deployers.stub.deployer as stub_deployer
        return stub_deployer.StubDeployDriver


fetcher_connector = _FetcherConnector()
driver_connector = _DriverConnector()


class Fetcher(abc.ABC):
    @abc.abstractmethod
    async def fetch_packages(self,
                             devices: list[Device],
                             processes: int = 1,
                             max_slots: int = 0,
                             ) -> tuple[dict[Device, str], dict[Device, Any]]:
        pass

    @abc.abstractmethod
    async def fetch(self,
                    devices: list[Device],
                    files_to_download: dict[str, list[str]] | None = None,
                    processes: int = 1,
                    max_slots: int = 0,
                    ):
        pass


def get_fetcher() -> Fetcher:
    connectors = fetcher_connector.get_all()
    fetcher, _ = get_connector_from_config("fetcher", connectors)
    return fetcher


class DeployDriver(abc.ABC):
    @abc.abstractmethod
    async def bulk_deploy(self, deploy_cmds: dict, args: DeployOptions, progress_bar: ProgressBar | None = None) -> DeployResult:
        pass

    @abc.abstractmethod
    def apply_deploy_rulebook(self, hw: HardwareView, cmd_paths, do_finalize=True, do_commit=True):
        pass

    @abc.abstractmethod
    def build_configuration_cmdlist(self, hw: HardwareView, do_finalize=True, do_commit=True):
        pass

    @abc.abstractmethod
    def build_exit_cmdlist(self, hw):
        pass


def get_deployer() -> DeployDriver:
    connectors = driver_connector.get_all()
    deployer, _ = get_connector_from_config("deployer", connectors)
    return deployer


# ===
def scrub_config(text: str, breed: str) -> str:
    return text


def show_bulk_report(hostnames, results, durations, log_dir):
    pass


class RulebookQuestionHandler:
    def __init__(self, dialogs):
        self._dialogs = dialogs

    def __call__(self, dev: Connector, cmd: Command, match_content: bytes):
        content = match_content.strip()
        content = content.decode()
        for matcher, answer in self._dialogs.items():
            if matcher(content):
                return Command(answer.text)

        get_logger().info("no answer in rulebook. dialogs=%s match_content=%s", self._dialogs, match_content)
        return None


def rb_question_to_question(q: MakeMessageMatcher, a: Answer) -> Question:  # TODO: drop MakeMessageMatcher
    text: str = q._text  # pylint: disable=protected-access
    answer: str = a.text
    is_regexp = False
    if text.startswith("/") and text.endswith("/"):
        is_regexp = True
        text = text[1:-1]
    res = Question(question=text, answer=answer, is_regexp=is_regexp, not_send_nl=not a.send_nl)
    return res


def make_cmd_params(rule: Dict[str, Any]) -> Dict[str, Any]:
    if rule:
        qa_handler = RulebookQuestionHandler(rule["attrs"]["dialogs"])
        qa_list: List[Question] = []
        for matcher, answer in qa_handler._dialogs.items():  # pylint: disable=protected-access
            qa_list.append(rb_question_to_question(matcher, answer))
        return {
            "questions": qa_list,
            "timeout": rule["attrs"]["timeout"],
        }
    return {
        "timeout": 30,
    }


def make_apply_commands(rule: dict, hw: HardwareView, do_commit: bool, do_finalize: bool, path: Optional[str] = None):
    apply_logic = rule["attrs"]["apply_logic"]
    before, after = apply_logic(hw, do_commit=do_commit, do_finalize=do_finalize, path=path)
    return before, after


def fill_cmd_params(rules: OrderedDict, cmd: Command):
    rule = deploying.match_deploy_rule(rules, (cmd.cmd,), {})
    if rule:
        cmd_params = make_cmd_params(rule)
        cmd.questions = cmd_params.get("questions", None)
        cmd.timeout = cmd_params["timeout"]


def apply_deploy_rulebook(hw: HardwareView, cmd_paths, do_finalize=True, do_commit=True):
    rules = get_rulebook(hw)["deploying"]
    cmds_with_apply = []
    for cmd_path, context in cmd_paths.items():
        rule = deploying.match_deploy_rule(rules, cmd_path, context)
        cmd_params = make_cmd_params(rule)
        before, after = make_apply_commands(rule, hw, do_commit, do_finalize)

        cmd = Command(cmd_path[-1], **cmd_params)
        # XXX более чистый способ передавать-мета инфу о команде
        cmd.level = len(cmd_path) - 1
        cmds_with_apply.append((cmd, before, after))

    def _key(item):
        _cmd, before, after = item
        return (tuple(cmd.cmd for cmd in before), tuple(cmd.cmd for cmd in after))

    cmdlist = CommandList()
    for _k, cmd_before_after in itertools.groupby(cmds_with_apply, key=_key):
        cmd_before_after = list(cmd_before_after)
        _, before, after = cmd_before_after[0]
        for c in before:
            c.level = 0
            fill_cmd_params(rules, c)
            cmdlist.add_cmd(c)
        for cmd, _before, _after in cmd_before_after:
            cmdlist.add_cmd(cmd)
        for c in after:
            c.level = 0
            fill_cmd_params(rules, c)
            cmdlist.add_cmd(c)
    return cmdlist



#!/usr/bin/env python3
import logging
import sys

import annet
from annet import argparse, cli, generators, hardware, lib, rulebook, diff


# =====
@lib.catch_ctrl_c
def main():
    annet.assert_python_version()
    parser = argparse.ArgParser()
    cli.fill_base_args(parser, annet.__name__, "configs/logging.yaml")
    rulebook.rulebook_provider_connector.set(rulebook.DefaultRulebookProvider)
    hardware.hardware_connector.set(hardware.AnnetHardwareProvider)
    diff.file_differ_connector.set(diff.UnifiedFileDiffer)

    parser.add_commands(parser.find_subcommands(cli.list_subcommands()))
    try:
        return parser.dispatch(pre_call=annet.init, add_help_command=True)
    except (generators.GeneratorError, annet.ExecError) as e:
        logging.error(e)
        return 1
    except BaseException as e:
        if formatted_output := getattr(e, "formatted_output", None):
            logging.error(formatted_output)
            return 1
        raise e


if __name__ == "__main__":
    sys.exit(main())



import functools
import importlib

from annet.annlib.rulebook import common  # pylint: disable=unused-import # noqa: F401,F403
from annet.annlib.rulebook.common import *  # pylint: disable=wildcard-import,unused-wildcard-import # noqa: F401,F403


@functools.lru_cache()
def import_rulebook_function(name):
    from . import rulebook_provider_connector

    index = name.rindex(".")
    for root in rulebook_provider_connector.get().get_root_modules():
        try:
            module = importlib.import_module(f"{root}.{name[:index]}", package=__name__.rsplit(".", 1)[0])
            return getattr(module, name[index + 1:])
        except ImportError:
            pass
    raise ImportError(f"Could not import {name}")



from annet.annlib.lib import uniq
from annet.annlib.rulebook import common
from annet.annlib.types import Op

from annet.rulebook.cisco.iface import is_in_channel, is_ip_cmd, is_vpn_cmd


def diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    for iface_row in old:
        _filter_channel_members(old[iface_row])
    for iface_row in new:
        _filter_channel_members(new[iface_row])

    for iface_row in uniq(old, new):
        iface_old = old.get(iface_row, {})
        iface_new = new.get(iface_row, {})
        iface_pre = diff_pre[iface_row]["subtree"]
        vpn_changed = False
        for (op, cmd, _, _) in common.default_diff(iface_old, iface_new, iface_pre, _pops):
            if op in {Op.ADDED, Op.REMOVED}:
                vpn_changed |= is_vpn_cmd(cmd)
                break

        if vpn_changed:
            for cmd in list(iface_old.keys()):
                if is_ip_cmd(cmd) and not is_vpn_cmd(cmd):
                    del iface_old[cmd]

        if _is_diff_removed_lag_from_lag_member(iface_old, iface_new):
            for cmd in list(iface_old.keys()):
                if _is_allowed_on_old_lag_memeber(cmd):
                    del iface_old[cmd]

    ret = common.default_diff(old, new, diff_pre, _pops)

    return ret

# ===

# Вырезает все команды не разрешенные
# на членах агрегата. В running-config
# листинге они наследуются от самого port-channel


def _filter_channel_members(tree):
    if any(is_in_channel(x) for x in tree):
        for cmd in list(tree.keys()):
            if not _is_allowed_on_channel(cmd):
                del tree[cmd]


def _is_allowed_on_channel(cmd_line):
    return cmd_line.startswith((
        "channel-group",
        "cdp",
        "description",
        "inherit",
        "ip port",
        "ipv6 port",
        "mac port",
        "lacp",
        "switchport host",
        "switchport",
        "shutdown",
        "rate-limit cpu",
        "snmp trap link-status",
        "mtu",
        "macsec",  # NOCDEV-9008
    ))

# ===

# При удалении lag, часть команд удаляется сначала с lag, тем самым удаляя с lag member,
# но рассчитанный diff этот нюанс не учитывает


def _is_diff_removed_lag_from_lag_member(old, new):
    """
    Проверяем, что в old есть признак lag member, а в new его нет. Следовательно порт выводится из lag.
    """
    if any(is_in_channel(x) for x in old) and not any(is_in_channel(x) for x in new):
        return True
    return False


def _is_allowed_on_old_lag_memeber(cmd_line):
    """
    Эти команды принудительно добавим на интерфейс после удаления его из lag
    """
    return cmd_line.startswith((
        "mtu",
    ))



from abc import ABC, abstractmethod
from collections.abc import Sequence, Iterable
from typing import Any, Literal
from itertools import chain

from annet.generators import PartialGenerator
from annet.rpl import PrefixMatchValue, MatchField, SingleCondition, RoutingPolicy
from .entities import IpPrefixList, PrefixListNameGenerator, JuniperPrefixListNameGenerator


class PrefixListFilterGenerator(PartialGenerator, ABC):
    TAGS = ["policy", "rpl", "routing"]

    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_prefix_lists(self, device: Any) -> Sequence[IpPrefixList]:
        raise NotImplementedError()

    # huawei
    def acl_huawei(self, _):
        return r"""
        ip ip-prefix
        ip ipv6-prefix
        """

    def _huawei_prefix_list(
            self,
            prefix_type: Literal["ipv6-prefix", "ip-prefix"],
            plist: IpPrefixList,
    ) -> Iterable[Sequence[str]]:
        for i, m in enumerate(plist.members):
            ge, le = m.or_longer
            yield (
                "ip",
                prefix_type,
                plist.name,
                f"index {i * 5 + 5}",
                "permit",
                str(m.prefix.network_address).upper(),
                str(m.prefix.prefixlen),
            ) + (
                ("greater-equal", str(ge)) if ge is not None else ()
            ) + (
                ("less-equal", str(le)) if le is not None else ()
            )

    def run_huawei(self, device: Any):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)

        name_generator = PrefixListNameGenerator(prefix_lists, policies)
        processed_names = set()
        for policy in policies:
            for statement in policy.statements:
                cond: SingleCondition[PrefixMatchValue]
                for cond in statement.match.find_all(MatchField.ip_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._huawei_prefix_list("ip-prefix", plist)
                        processed_names.add(plist.name)
                for cond in statement.match.find_all(MatchField.ipv6_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._huawei_prefix_list("ipv6-prefix", plist)
                        processed_names.add(plist.name)

    # arista
    def acl_arista(self, _):
        return r"""
        ip prefix-list
            seq
        ipv6 prefix-list
            seq
        """

    def _arista_prefix_list(
            self,
            prefix_type: Literal["ipv6", "ip"],
            plist: IpPrefixList,
    ) -> Iterable[Sequence[str]]:
        with self.block(prefix_type, "prefix-list", plist.name):
            for i, m in enumerate(plist.members):
                ge, le = m.or_longer
                yield (
                    f"seq {i * 10 + 10}",
                    "permit",
                    str(m.prefix),
                ) + (
                    ("ge", str(ge)) if ge is not None else ()
                ) + (
                    ("le", str(le)) if le is not None else ()
                )

    def run_arista(self, device: Any):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)
        name_generator = PrefixListNameGenerator(prefix_lists, policies)
        processed_names = set()
        for policy in policies:
            for statement in policy.statements:
                cond: SingleCondition[PrefixMatchValue]
                for cond in statement.match.find_all(MatchField.ip_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._arista_prefix_list("ip", plist)
                        processed_names.add(plist.name)
                for cond in statement.match.find_all(MatchField.ipv6_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._arista_prefix_list("ipv6", plist)
                        processed_names.add(plist.name)

    # Cisco IOS XR
    def acl_iosxr(self, device: Any):
        return r"""
        prefix-set
            ~ %global=1
        """

    def _iosxr_prefixlist(self, prefixlist: IpPrefixList):
        with self.block("prefix-set", prefixlist.name):
            for n, member in enumerate(prefixlist.members):
                if n + 1 < len(prefixlist.members):
                    comma = ","
                else:
                    comma = ""

                ge, le = member.or_longer
                if ge is le is None:
                    yield f"{member.prefix}{comma}",
                elif ge is None:
                    yield f"{member.prefix} le {le}{comma}",
                elif le is None:
                    yield f"{member.prefix} ge {ge}{comma}",
                elif ge == le:
                    yield f"{member.prefix} eq {ge}{comma}",
                else:
                    yield f"{member.prefix} ge {ge} le {le}{comma}",

    def run_iosxr(self, device: Any):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)

        name_generator = PrefixListNameGenerator(prefix_lists, policies)
        processed_names = set()
        for policy in policies:
            for statement in policy.statements:
                cond: SingleCondition[PrefixMatchValue]
                for cond in statement.match.find_all(MatchField.ip_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._iosxr_prefixlist(plist)
                        processed_names.add(plist.name)
                for cond in statement.match.find_all(MatchField.ipv6_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._iosxr_prefixlist(plist)
                        processed_names.add(plist.name)

    def acl_juniper(self, _):
        return r"""
        policy-options     %cant_delete
            prefix-list *
                ~
            route-filter-list *
                ~
        """

    def _juniper_prefixlist(self, name: str, prefixlist: IpPrefixList):
        with self.block("policy-options"):
            with self.block("prefix-list", name):
                for member in prefixlist.members:
                    yield f"{member.prefix}"

    def _juniper_router_filter_list(self, name: str, prefixlist: IpPrefixList):
        with self.block("policy-options"):
            with self.block("route-filter-list", name):
                for member in prefixlist.members:
                    ge, le = member.or_longer
                    if ge is None and le is None:
                        yield f"{member.prefix} exact"
                        continue
                    if ge is None:
                        ge = member.prefix.prefixlen
                    if le is None:
                        le = member.prefix.max_prefixlen
                    # may produce config that is not accepted by commit
                    # since juniper enforces that n <= ge <= le
                    # where n is prefix len: .../n

                    # this is done specifically to match other generators behaviour
                    # can be revised in two ways: exeption or enforce via max/min
                    # but need to be consistent across vendors so will leave it for now
                    yield f"{member.prefix}", "prefix-length-range", f"/{ge}-/{le}"

    def run_juniper(self, device: Any):
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)

        name_generator = JuniperPrefixListNameGenerator(prefix_lists, policies)
        processed_names: set[str] = set()
        for policy in policies:
            for statement in policy.statements:
                conds = chain(
                    statement.match.find_all(MatchField.ip_prefix),
                    statement.match.find_all(MatchField.ipv6_prefix),
                )
                cond: SingleCondition[PrefixMatchValue]
                for cond in conds:
                    for cond_name in cond.value.names:
                        plist = name_generator.get_prefix(cond_name, cond.value)
                        plist_type = name_generator.get_type(cond_name, cond.value)

                        if plist.name not in processed_names:
                            processed_names.add(plist.name)
                            if plist_type == "prefix-list":
                                yield from self._juniper_prefixlist(plist.name, plist)
                            elif plist_type == "route-filter":
                                yield from self._juniper_router_filter_list(plist.name, plist)
                            else:
                                raise NotImplementedError(
                                    f"Prefix list {cond_name} type {plist_type} is not supported for Juniper",
                                )



from typing import Optional

from annet.adapters.netbox.common.models import NetboxDevice
from annet.rpl import R, RouteMap, Route

routemap = RouteMap[NetboxDevice]()


def find_loopback(device: NetboxDevice) -> Optional[str]:
    for iface in device.interfaces:
        if iface.name.lower().startswith("lo"):
            return iface.name
    return None


SOME_CONDITION = (R.rd.has_any("RD_EXAMPLE1")) & (R.protocol == "bgp")


@routemap
def example1(device: NetboxDevice, route: Route):
    # condition can be referenced as global constant
    with route(SOME_CONDITION, number=1, name="n1") as rule:
        rule.set_local_pref(100)
        rule.set_metric(100)
        rule.add_metric(200)
        rule.community.set("COMMUNITY_EXAMPLE_ADD")
        rule.as_path.set(12345, "123456")
        rule.allow()

    loopback = find_loopback(device)
    if loopback:  # rules can be generated dynamically
        with route(
                R.community.has("COMMUNITY_EXAMPLE_REMOVE"),
                R.interface == loopback,  # conditions can reference calculated values
                number=2, name="n2",
        ) as rule:
            rule.set_local_pref(100)
            rule.add_metric(200)
            rule.community.add("COMMUNITY_EXAMPLE_ADD")
            rule.community.remove("COMMUNITY_EXAMPLE_REMOVE")
            rule.allow()
    # default rule with no conditions
    with route(number=100, name="last") as rule:
        rule.deny()


@routemap
def example2(device: NetboxDevice, route: Route):
    with route(R.as_path_filter("ASP_EXAMPLE"), number=3, name="n3") as rule:
        rule.deny()
    with route(R.match_v6("IPV6_LIST_EXAMPLE"), number=4, name="n4") as rule:
        rule.allow()
    with route(R.match_v4("IPV4_LIST_EXAMPLE", or_longer=(29, 48)), number=5, name="n5") as rule:
        rule.allow()

    with route(R.as_path_length >= 1, R.as_path_length <= 20, number=6, name="n6") as rule:
        rule.allow()
    with route(R.extcommunity_rt.has("EXTCOMMUNITY_EXAMPLE_REMOVE"), number=7, name="n7") as rule:
        rule.extcommunity_rt.remove("EXTCOMMUNITY_EXAMPLE_REMOVE")
        rule.deny()


# this policy is not used in mesh model, so will be filtered out
@routemap
def unused(device: NetboxDevice, route: Route):
    with route(number=3, name="n3") as rule:
        rule.deny()



from typing import Annotated, Optional, Union

from annet.bgp_models import Family, Redistribute, Aggregate
from .basemodel import BaseMeshModel, Concat, DictMerge, Merge, KeyDefaultDict
from .peer_models import MeshPeerGroup


class _Aggregate(BaseMeshModel):
    policy: str
    routes: Annotated[tuple[str, ...], Concat()]
    as_path: str
    reference: str
    suppress: bool
    discard: bool
    as_set: bool


class FamilyOptions(BaseMeshModel):
    def __init__(self, **kwargs):
        kwargs.setdefault("aggregate", _Aggregate())
        super().__init__(**kwargs)
    family: Family
    vrf_name: str
    multipath: int
    global_multipath: int
    aggregate: Annotated[_Aggregate, Merge()]  # use `aggregates` instead
    aggregates: Annotated[tuple[Aggregate, ...], Concat()]
    af_loops: int
    redistributes: Annotated[tuple[Redistribute, ...], Concat()]
    allow_default: bool
    aspath_relax: bool
    igp_ignore: bool
    next_hop_policy: bool
    rib_import_policy: bool
    advertise_l2vpn_evpn: bool
    rib_group: bool
    loops: int
    advertise_bgp_static: bool
    import_policy: Optional[str]
    export_policy: Optional[str]


class _FamiliesMixin:
    def __init__(self, **kwargs):
        kwargs.setdefault("ipv4_unicast", FamilyOptions(family="ipv4_unicast"))
        kwargs.setdefault("ipv6_unicast", FamilyOptions(family="ipv6_unicast"))
        kwargs.setdefault("ipv4_vpn_unicast", FamilyOptions(family="ipv4_vpn_unicast"))
        kwargs.setdefault("ipv6_vpn_unicast", FamilyOptions(family="ipv6_vpn_unicast"))
        kwargs.setdefault("ipv4_labeled_unicast", FamilyOptions(family="ipv4_labeled_unicast"))
        kwargs.setdefault("ipv6_labeled_unicast", FamilyOptions(family="ipv6_labeled_unicast"))
        kwargs.setdefault("l2vpn_evpn", FamilyOptions(family="l2vpn_evpn"))
        super().__init__(**kwargs)
    ipv4_unicast: Annotated[FamilyOptions, Merge()]
    ipv6_unicast: Annotated[FamilyOptions, Merge()]
    ipv4_vpn_unicast: Annotated[FamilyOptions, Merge()]
    ipv6_vpn_unicast: Annotated[FamilyOptions, Merge()]
    ipv4_labeled_unicast: Annotated[FamilyOptions, Merge()]
    ipv6_labeled_unicast: Annotated[FamilyOptions, Merge()]
    l2vpn_evpn: Annotated[FamilyOptions, Merge()]


class VrfOptions(_FamiliesMixin, BaseMeshModel):
    def __init__(self, vrf_name: str, **kwargs):
        kwargs.setdefault("ipv4_unicast", FamilyOptions(family="ipv4_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv6_unicast", FamilyOptions(family="ipv6_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv4_vpn_unicast", FamilyOptions(family="ipv4_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv6_vpn_unicast", FamilyOptions(family="ipv6_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv4_labeled_unicast", FamilyOptions(family="ipv4_labeled_unicast", vrf_name=vrf_name))
        kwargs.setdefault("ipv6_labeled_unicast", FamilyOptions(family="ipv6_labeled_unicast", vrf_name=vrf_name))
        kwargs.setdefault("l2vpn_evpn", FamilyOptions(family="l2vpn_evpn", vrf_name=vrf_name))
        kwargs.setdefault("groups", KeyDefaultDict(lambda x: MeshPeerGroup(name=x)))
        super().__init__(vrf_name=vrf_name, **kwargs)

    vrf_name: str
    vrf_name_global: Optional[str]
    as_path_relax: bool
    import_policy: Optional[str]
    export_policy: Optional[str]
    l3vni: Optional[int]
    rt_import: Annotated[tuple[str, ...], Concat()]
    rt_export: Annotated[tuple[str, ...], Concat()]
    rt_import_v4: Annotated[tuple[str, ...], Concat()]
    rt_export_v4: Annotated[tuple[str, ...], Concat()]
    route_distinguisher: Optional[str]
    static_label: Optional[int]  # FIXME: str?
    groups: Annotated[dict[str, MeshPeerGroup], DictMerge(Merge())]


class L2VpnOptions(BaseMeshModel):
    name: str
    vid: str | int  # VLAN ID, possible values are 1 to 4094, ranges can be set as strings
    l2vni: int  # VNI, possible values are 1 to 2**24-1
    route_distinguisher: str  # like in VrfOptions
    rt_import: Annotated[tuple[str, ...], Concat()]  # like in VrfOptions
    rt_export: Annotated[tuple[str, ...], Concat()]  # like in VrfOptions
    advertise_host_routes: bool  # advertise IP+MAC routes into L3VNI


class GlobalOptionsDTO(_FamiliesMixin, BaseMeshModel):
    def __init__(self, **kwargs):
        kwargs.setdefault("groups", KeyDefaultDict(lambda x: MeshPeerGroup(name=x)))
        kwargs.setdefault("vrf", KeyDefaultDict(lambda x: VrfOptions(vrf_name=x)))
        kwargs.setdefault("l2vpn", KeyDefaultDict(lambda x: L2VpnOptions(name=x)))
        super().__init__(**kwargs)

    as_path_relax: bool
    local_as: Union[int, str]
    loops: int
    multipath: int
    router_id: str
    cluster_id: Optional[str]
    vrf: Annotated[dict[str, VrfOptions], DictMerge(Merge())]
    groups: Annotated[dict[str, MeshPeerGroup], DictMerge(Merge())]
    l2vpn: Annotated[dict[str, L2VpnOptions], DictMerge(Merge())]



import ipaddress
from annet.annlib.types import Op
from annet.rulebook import common
from annet.rulebook.common import DiffItem


def ipv6_addr(old, new, diff_pre, _pops):
    """
    Приводим все ipv6-адреса в объекты IPv6Interface и далее сравниваем
    """
    address_new_line = [a for a in map(_parse_ipv6, new) if a]
    address_old_line = [a for a in map(_parse_ipv6, old) if a]

    ret = []
    for item in common.default_diff(old, new, diff_pre, _pops):
        # Проверяем адрес помеченный под снос на наличии в новом списке
        if item.op == Op.REMOVED and _parse_ipv6(item.row) in address_new_line:
            result_item = DiffItem(Op.AFFECTED, item.row, item.children, item.diff_pre)
        # Проверяем адрес помеченный для добавления на наличии в старом списке
        elif item.op == Op.ADDED and _parse_ipv6(item.row) in address_old_line:
            result_item = None
        # Остальное без изменений
        else:
            result_item = item
        if result_item:
            ret.append(result_item)
    return ret


def _parse_ipv6(row):
    """
    Парсит IPv6-интерфейс из строки, предполагая, что адрес находится на втором месте.
    Возвращает объект IPv6Interface или None.
    """
    if row:
        parts = row.split()
        if len(parts) > 1:
            return ipaddress.IPv6Interface(parts[1])
    return None



from collections.abc import Iterator, Iterable
from dataclasses import dataclass
from enum import Enum
from typing import Generic, TypeVar, Any


class ActionType(Enum):
    SET = "set"
    ADD = "add"
    REMOVE = "delete"

    CUSTOM = "custom"


ValueT = TypeVar("ValueT")


@dataclass
class SingleAction(Generic[ValueT]):
    field: str
    type: ActionType
    value: ValueT


class Action:
    def __init__(self) -> None:
        self.actions: list[SingleAction[Any]] = []

    def append(self, action: SingleAction[Any]) -> None:
        self.actions.append(action)

    def __repr__(self):
        actions = ", ".join(repr(c) for c in self.actions)
        return f"Action({actions})"

    def __getitem__(self, item: str) -> SingleAction[Any]:
        return next(c for c in self.actions if c.field == item)

    def __contains__(self, item: str) -> bool:
        return any(c.field == item for c in self.actions)

    def __len__(self) -> int:
        return len(self.actions)

    def __iter__(self) -> Iterator[SingleAction[Any]]:
        return iter(self.actions)

    def find_all(self, item: str) -> Iterable[SingleAction[Any]]:
        for action in self.actions:
            if action.field == item:
                yield action



import abc
import argparse
import contextlib
import functools
import os
import sys
from functools import lru_cache
from typing import (
    Callable,
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from annet.connectors import Connector


T = TypeVar("T")

_ARGS_ATTR_NAME = "_alan_opts"


# =====
class FuncMeta:
    opts: List[Callable]
    parent: Optional[Callable]
    parser: Optional[argparse.ArgumentParser]

    def __init__(self):
        self.cmd_name = ""
        self.opts = []
        self.parent = None
        self.parser = None
        self.sub = None

    def __repr__(self):
        return f"{self.__class__.__name__}(" + ", ".join((
            f"cmd_name={self.cmd_name}",
            f"opt={self.opts}",
            f"parent={self.parent}",
            f"parser={self.parser}",
            f"sub={self.sub}",
        )) + ")"


def _get_meta(func: Callable) -> FuncMeta:
    if not hasattr(func, _ARGS_ATTR_NAME):
        _reset_meta(func)
    return getattr(func, _ARGS_ATTR_NAME)


def _reset_meta(func: Callable):
    setattr(func, _ARGS_ATTR_NAME, FuncMeta())


class ConvertibleDefault:
    def __init__(self, value: str):
        self.value = value

    def convert(self, type_function: Callable[[str], T]) -> T:
        return type_function(self.value)


class DefaultFromEnv(ConvertibleDefault):
    def __init__(self, var_name: str, default_val: str):
        super().__init__(os.environ.get(var_name, default_val))
        self.var_name = var_name

    def convert(self, type_function: Callable[[str], T]) -> T:
        try:
            return super().convert(type_function)
        except Exception:
            function_name = getattr(type_function, "__name__", repr(type_function))
            raise ValueError(f"invalid {function_name} in the {self.var_name} environment variable: {repr(self.value)}")


class Arg:
    """аргумент CLI"""

    def __init__(self, *args, **kwargs):
        """Конструктор повторяет прототип parser.add_argument()"""
        if args and isinstance(args[0], type(self)):
            # copy constructor
            args = args[0].args
            new_kwargs = args[0].kwargs.copy()
            new_kwargs.update(kwargs)
            kwargs = new_kwargs

        self.args = args
        self.kwargs = kwargs

        self._prepared = False
        self.default = None

        self.dest = None  # заполняется в attach()

    def _prepare(self):
        if self._prepared:
            return
        self._prepared = True
        default = self.kwargs.get("default", None)
        if isinstance(default, ConvertibleDefault) and "type" in self.kwargs:
            default = self.kwargs["default"] = default.convert(self.kwargs["type"])
        elif isinstance(default, Callable):
            default = self.kwargs["default"] = default()
        elif default is False and "action" not in self.kwargs:
            self.kwargs["action"] = "store_true"
        self.default = default

        if default is not None and not (default is False and self.kwargs.get("action") == "store_true"):
            if "help" not in self.kwargs:
                self.kwargs["help"] = ""
            self.kwargs["help"] += " (default: %r)" % default

    def attach(self, parser: argparse.ArgumentParser):
        self._prepare()
        arg = parser.add_argument(*self.args, **self.kwargs)
        self.dest = arg.dest


class ArgGroup:
    """
    Контейнер для нескольких аргументов.
    Класс служит для описания набора аргументов, а экземляр - для хранения значений.

    Пример:
    class Group1(ArgGroup):
        in = Arg("--in")
        out = Arg("--out")

    values = Group1(in="/dev/null", out="/dev/null")
    open(values.in)
    """
    def __init__(self, *args, **kwargs):
        """
        В kwargs - пары ключ-значение. Соотвествующие опции должны быть объявлены в классе
        """
        keys = {arg_name: arg.default for arg_name, arg in self._enum_args().items()}
        for src_obj in args:
            for opt, val in vars(src_obj).items():
                if opt in keys:
                    keys[opt] = val
        for opt, val in kwargs.items():
            if opt in keys:
                keys[opt] = val
            else:
                raise TypeError("Unknown argument %r for %s" % (opt, type(self).__qualname__))
        self.__dict__.update(keys)

    @classmethod
    @lru_cache()
    def _enum_args(cls) -> Dict[str, Arg]:
        ret = {}
        for base in cls.__mro__:
            for name, value in vars(base).items():
                if not name.startswith("_") and isinstance(value, Arg):
                    ret[name] = value
        return ret

    @classmethod
    def attach(cls, parser: argparse.ArgumentParser):
        for arg in cls._enum_args().values():
            arg.attach(parser)

    @classmethod
    def construct_from(cls, ns: argparse.Namespace) -> "ArgGroup":
        kwargs = {}
        for arg_name, arg in cls._enum_args().items():
            kwargs[arg_name] = getattr(ns, arg.dest)
        return cls(**kwargs)

    @classmethod
    def copy_from(cls, other: "ArgGroup", **more_kwargs) -> "ArgGroup":
        kwargs = {}
        for arg_name in cls._enum_args():
            try:
                kwargs[arg_name] = getattr(other, arg_name)
            except AttributeError:
                pass

        kwargs.update(more_kwargs)
        return cls(**kwargs)

    def __repr__(self):
        return "%(classname)s(%(params)s)" % dict(
            classname=type(self).__qualname__,
            params=", ".join(
                "%s=%r" % (key, getattr(self, key))
                for key in sorted(self._enum_args())
            )
        )

    def stdin(self, **kwargs):
        ret = {}
        stdin_used = False
        for arg, val in kwargs.items():
            if val == "-":
                if stdin_used:
                    raise ValueError("stdin can not be used twice")
                self.validate_stdin(arg, val, **kwargs.copy())
                ret[arg] = _read_stdin_once()
                stdin_used = True
            else:
                ret[arg] = None
        return ret

    def validate_stdin(self, arg, val, **kwargs):
        pass


class _HelpFormatter(argparse.HelpFormatter):
    def __init__(self, *args, **kwargs):
        kwargs["width"] = self._get_term_width()
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_term_width():
        try:
            return int(os.environ["COLUMNS"])
        except (KeyError, ValueError):
            try:
                return os.get_terminal_size(sys.stdout.fileno()).columns
            except OSError:
                return None


class ArgDispatcher(abc.ABC):
    @abc.abstractmethod
    def setup(self) -> ContextManager[None]:
        pass

    @abc.abstractmethod
    def exec(self) -> ContextManager[None]:
        pass

    @abc.abstractmethod
    def pre_call(self) -> ContextManager[None]:
        pass

    @contextlib.contextmanager
    def func_opts(self, arg: Union[Arg, Type[ArgGroup]]) -> ContextManager[None]:
        pass

    @abc.abstractmethod
    def func(self, ns: argparse.Namespace) -> ContextManager[None]:
        pass


class NullArgDispatcher(ArgDispatcher):
    @contextlib.contextmanager
    def setup(self) -> Iterator[None]:
        yield

    @contextlib.contextmanager
    def exec(self) -> ContextManager[None]:
        yield

    @contextlib.contextmanager
    def pre_call(self) -> ContextManager[None]:
        yield

    @contextlib.contextmanager
    def func_opts(self, arg: Union[Arg, Type[ArgGroup]]) -> ContextManager[None]:
        yield

    @contextlib.contextmanager
    def func(self, ns: argparse.Namespace) -> ContextManager[None]:
        yield


class _DispatcherConnector(Connector[ArgDispatcher]):
    name = "Dispatcher"
    ep_name = "dispatcher"

    def _get_default(self) -> Type["ArgDispatcher"]:
        return NullArgDispatcher


dispatcher_connector = _DispatcherConnector()


class ArgParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        if "formatter_class" not in kwargs:
            kwargs["formatter_class"] = _HelpFormatter
        super().__init__(*args, **kwargs)
        self._dispatcher: ArgDispatcher = dispatcher_connector.get()

    def argv(self):
        return sys.argv[1:]

    def add_commands(self, commands: Iterable[Callable]):
        subparsers = self.add_subparsers()
        commands = list(commands)
        failcount = 0
        while commands and failcount < len(commands):
            func, commands = commands[0], commands[1:]
            if not self.add_func(func, subparsers):
                commands.append(func)
                failcount += 1
            else:
                failcount = 0
        if failcount:
            raise RuntimeError("Failed to resolve subparsers")

    def add_func(self, func, sub):
        meta = _get_meta(func)
        parent = meta.parent
        if parent:
            parent_meta = _get_meta(parent)
            if not parent_meta.parser:
                return False
            if not parent_meta.sub:
                parent_meta.sub = parent_meta.parser.add_subparsers()
            sub = parent_meta.sub

        sp = sub.add_parser(meta.cmd_name, help=func.__doc__)
        sp.set_defaults(func=func)
        meta.parser = sp

        for arg in self.func_opts(func):
            arg.attach(sp)
        return True

    def set_dispatcher(self, dispatcher: ArgDispatcher) -> None:
        self._dispatcher = dispatcher

    def dispatch(self, pre_call=None, add_help_command=False):
        with self._dispatcher.setup():
            with self._dispatcher.exec():
                argv = self.argv()
                if add_help_command and argv and argv[0] == "help":
                    argv.pop(0)
                    argv.append("--help")

                # show usage in cache when no options specified
                if not argv:
                    argv = [""]

                ns = self.parse_args(argv)
                assert ns.func
                if pre_call:
                    with self._dispatcher.pre_call():
                        pre_call(ns)

                values = []
                for arg in self.func_opts(ns.func):
                    with self._dispatcher.func_opts(arg):
                        if isinstance(arg, Arg):
                            values.append(getattr(ns, arg.dest))
                        else:
                            # arg is a ArgGroup-derived type
                            values.append(arg.construct_from(ns))

                with self._dispatcher.func(ns):
                    return ns.func(*values)

    @staticmethod
    def find_subcommands(variables):
        for var in variables.values():
            if callable(var) and hasattr(var, _ARGS_ATTR_NAME):
                yield var

    @classmethod
    def func_opts(cls, func):
        yield from _get_meta(func).opts


def subcommand(*arg_list: Union[Arg, Type[ArgGroup]], parent: Callable = None, is_group: bool = False):
    """
    декоратор, задающий cli-аргументы подпрограммы

    Пример:
    @subcommand(Arg(), ArgGroup)
    def cmd1(arg1, arg2):
        pass

    Связь аргументов происходит только по порядковому номеру, каждый аргумент subcommand становится аргументом функции.
    Функция вызывается только с позиционными аргументами, всегда с одним и тем же количеством аргументов.

    Для создания более одного уровня команд используется аргумент parent

    Пример: 'ann some thing' - вызовет some_thing()

    @subcommand()
    def some():
        pass


    @subcommand(parent=some)
    def some_thing():
        pass
    """
    def _amend_func(func):
        meta = _get_meta(func)
        if is_group:
            @functools.wraps(func)
            def func():
                meta.parser.print_help()
        cmd_name = func.__name__
        if parent:
            parentprefix = parent.__name__ + "_"
            if cmd_name.startswith(parentprefix):
                cmd_name = cmd_name[len(parentprefix):]
            meta.parent = parent

        meta.cmd_name = cmd_name.replace("_", "-").lower()
        meta.opts.extend(arg_list)
        return func
    return _amend_func


@functools.lru_cache()
def _read_stdin_once():
    return sys.stdin.read()



from annet.deploy import Fetcher
from annet.connectors import AdapterWithConfig
from typing import Dict, List, Any
from annet.storage import Device


class StubFetcher(Fetcher, AdapterWithConfig):
    @classmethod
    def with_config(cls, **kwargs: Dict[str, Any]) -> Fetcher:
        return cls(**kwargs)

    async def fetch_packages(self,
                             devices: list[Device],
                             processes: int = 1,
                             max_slots: int = 0,
                             ) -> tuple[dict[Device, str], dict[Device, Any]]:
        raise NotImplementedError()

    async def fetch(self,
                    devices: list[Device],
                    files_to_download: dict[str, list[str]] | None = None,
                    processes: int = 1,
                    max_slots: int = 0,
                    ):
        raise NotImplementedError()



import functools
import re
from collections import OrderedDict as odict

from annet.annlib.lib import jun_activate, jun_is_inactive, merge_dicts
from annet.vendors.tabparser import JuniperFormatter
from annet.annlib.types import Op
from annet.rulebook import common


def comment_processor(item: common.DiffItem):
    if item.op in (Op.REMOVED, Op.ADDED) and item.row.startswith(JuniperFormatter.Comment.begin):
        comment = JuniperFormatter.Comment.loads(item.row)

        item.diff_pre["attrs"]["context"]["comment"] = True
        item.diff_pre["attrs"]["context"]["row"] = comment.row
        item.diff_pre["key"] = item.diff_pre["raw_rule"] = (
            f"{JuniperFormatter.Comment.begin} {comment.row} {JuniperFormatter.Comment.end}"
        )

        return common.DiffItem(
            item.op,
            f"{JuniperFormatter.Comment.begin} {comment.comment} {JuniperFormatter.Comment.end}",
            item.children,
            item.diff_pre,
        )
    else:
        return item


def inactive_blocks(diff_foo):
    @functools.wraps(diff_foo)
    def wrapper(old: odict, new: odict, diff_pre, *args, **kwargs):
        old_inactives = list(map(jun_activate, filter(jun_is_inactive, old)))
        new_inactives = list(map(jun_activate, filter(jun_is_inactive, new)))

        diff: list[common.DiffItem]

        if len(old_inactives) == 0 and len(new_inactives) == 0:
            diff = diff_foo(old, new, diff_pre, *args, **kwargs)
        else:
            inactive_pre = odict([(jun_activate(k), v) for k, v in diff_pre.items() if jun_is_inactive(k)])
            merged_pre = merge_dicts(diff_pre, inactive_pre)

            diff = diff_foo(
                strip_toplevel_inactives(old),
                strip_toplevel_inactives(new),
                merged_pre,
                *args, **kwargs
            )

            for activated in [k for k in old_inactives if k in new]:
                diff.append(
                    common.DiffItem(
                        Op.ADDED,
                        activate_cmd(activated, merged_pre),
                        [],
                        diff_pre[activated]["match"]
                    )
                )

            for deactivated in [k for k in new_inactives if k not in old_inactives]:
                # если деактивуруемого блока не существует - ставим один deactivate, глубже не идем
                if deactivated not in diff_pre:
                    diff = [
                        common.DiffItem(
                            Op.ADDED,
                            deactivate_cmd(deactivated, merged_pre),
                            [],
                            inactive_pre[deactivated]["match"]
                        )
                    ]
                else:
                    diff.append(
                        common.DiffItem(
                            Op.ADDED,
                            deactivate_cmd(deactivated, merged_pre),
                            [],
                            diff_pre[deactivated]["match"]
                        )
                    )

        return list(map(comment_processor, diff))

    return wrapper


@inactive_blocks
def default_diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    diff = common.default_diff(old, new, diff_pre, _pops)
    diff = ignore_quotes(diff)
    diff = strip_inactive_removed(diff)
    return diff


@inactive_blocks
def ordered_diff(old, new, diff_pre, _pops=(Op.AFFECTED,)):
    diff = common.ordered_diff(old, new, diff_pre, _pops)
    diff = ignore_quotes(diff)
    diff = strip_inactive_removed(diff)
    return diff


# =====
def strip_toplevel_inactives(tree):
    for inactive in filter(jun_is_inactive, tree):
        assert jun_activate(inactive) not in tree
    return odict([(k, v) if not jun_is_inactive(k) else (jun_activate(k), v) for k, v in tree.items()])


def activate_cmd(active_key, diff_pre):
    return cmd(active_key, diff_pre, "activate")


def deactivate_cmd(active_key, diff_pre):
    return cmd(active_key, diff_pre, "deactivate")


def cmd(active_key, diff_pre, cmd):
    assert not jun_is_inactive(active_key)
    if not diff_pre[active_key]["subtree"]:
        # Если конанда не имеет подблоков И имеет агрументы то надо их отбросить
        return " ".join([cmd, active_key.split()[0]])
    return " ".join([cmd, active_key])


def ignore_quotes(diff):
    """
    Фильтрует из diff строки которые различаются
    только наличием/отсутсвием кавычек
    i.e.
    description "loopbacks";
    description loopbacks;
    эквивалентны
    """
    equivs = {}
    for elem in diff:
        key = strip_quotes(elem[1])
        if key not in equivs:
            equivs[key] = 0
        equivs[key] += 1
    filtered_diff = [elem for elem in diff if equivs[strip_quotes(elem[1])] == 1]
    return filtered_diff


def strip_quotes(key):
    return re.sub(r"\"(?P<quoted_text>[^\"]+)\"$", r"\g<quoted_text>", key)


def strip_inactive_removed(diff):
    for elem in diff:
        if elem[0] == Op.REMOVED and elem[3]["key"]:
            key = elem[3]["key"][0]
            if jun_is_inactive(key):
                elem[3]["key"] = tuple([jun_activate(key)])
    return diff



#!/usr/bin/env python3
import os
import setuptools

here = os.path.abspath(os.path.dirname(__file__))


def requirements() -> str:
    with open(os.path.join(here, "requirements.txt")) as file:
        return "".join(
            line
            for line in file.readlines()
            if not line.startswith("-")
        )


if __name__ == "__main__":
    setuptools.setup(
        name="annet",
        version=os.getenv("VERSION") or "0.0",
        description="annet",
        license="MIT",
        url="https://github.com/annetutil/annet",
        packages=setuptools.find_packages(include=[
            "annet",
            "annet.*",
            "annet_generators",
            "annet_generators.*",
        ]),
        package_data={
            "annet": ["configs/*"],
            "annet.rulebook": ["texts/*.rul", "texts/*.order", "texts/*.deploy"],
            "annet.annlib.netdev.devdb": ["data/*.json"],
        },
        entry_points={
          "console_scripts": [
              "annet = annet.annet:main",
          ],
          "annet.connectors": [
            "storage = annet.adapters.netbox.provider:NetboxProvider",
          ],
          "annet.connectors.storage": [
            "file = annet.adapters.file.provider:Provider",
          ],
        },
        extras_require={
            "netbox": ["annetbox[sync]>=0.7.0"],
        },
        python_requires=">=3.10",
        install_requires=requirements(),
        include_package_data=True,
    )



from dataclasses import dataclass
from typing import Dict
import re

from adaptix import Retort, name_mapping, NameStyle
from dataclass_rest import get
from dataclass_rest.client_protocol import FactoryProtocol

from .client import BaseNetboxClient


@dataclass
class Status:
    netbox_version: str
    plugins: Dict[str, str]

    @property
    def minor_version(self) -> str:
        if match := re.match(r"\d+\.\d+", self.netbox_version):
            return match.group(0)
        return ""


class NetboxStatusClient(BaseNetboxClient):
    def _init_response_body_factory(self) -> FactoryProtocol:
        return Retort(recipe=[
            name_mapping(name_style=NameStyle.LOWER_KEBAB)
        ])

    @get("status/")
    def status(self) -> Status:
        ...



import time
from typing import Optional, Union

from annet import tracing
from annet.tracing import tracing_connector
from annet.types import GeneratorPartialRunArgs, GeneratorPerf

from .base import BaseGenerator


class GeneratorPerfMesurer:
    def __init__(
        self,
        gen: BaseGenerator,
        run_args: Optional[GeneratorPartialRunArgs] = None,
        trace_min_duration: tracing.MinDurationT = None
    ):
        self._gen = gen
        self._run_args = run_args

        self._start_time: float = 0.0
        self._span_ctx = None
        self._span = None
        self._trace_min_duration = trace_min_duration

        self.last_result: Optional[GeneratorPerf] = None

    def start(self) -> None:
        self.last_result = None

        self._span_ctx = tracing_connector.get().start_as_current_span(
            "gen:call",
            tracer_name=self._gen.__class__.__module__,
            min_duration=self._trace_min_duration,
        )
        self._span = self._span_ctx.__enter__()  # pylint: disable=unnecessary-dunder-call

        if self._span:
            self._span.set_attributes({"generator.class": self._gen.__class__.__name__})
            if self._run_args:
                tracing_connector.get().set_device_attributes(
                    self._span, self._run_args.device,
                )

        self._start_time = time.monotonic()

    def finish(self, exc_type=None, exc_val=None,
               exc_tb=None) -> GeneratorPerf:
        total = time.monotonic() - self._start_time
        self._span_ctx.__exit__(exc_type, exc_val, exc_tb)
        rt = self._gen.storage.flush_perf()

        meta = {}
        if tracing_connector.get().enabled:
            span_context = self._span.get_span_context()
            meta = {
                "span": {
                    "trace_id": str(span_context.trace_id),
                    "span_id": str(span_context.span_id),
                },
            }

        self.last_result = GeneratorPerf(total=total, rt=rt, meta=meta)
        return self.last_result

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish(exc_type, exc_val, exc_tb)



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional

from annet.adapters.netbox.common.models import (
    IpAddress, NetboxDevice, Entity, Prefix, InterfaceType, Interface,
    IpFamily, Label, FHRPGroupAssignment, DeviceIp, FHRPGroup,
)


@dataclass
class DeviceIpV37(DeviceIp):
    family: int


@dataclass
class PrefixV37(Prefix):
    site: Optional[Entity] = None


@dataclass
class IpAddressV37(IpAddress[PrefixV37]):
    role: Optional[Label] = None
    prefix: Optional[PrefixV37] = None


@dataclass
class FHRPGroupV37(FHRPGroup[DeviceIpV37]):
    pass


@dataclass
class FHRPGroupAssignmentV37(FHRPGroupAssignment[FHRPGroupV37]):
    pass


@dataclass
class InterfaceV37(Interface[IpAddressV37, FHRPGroupAssignmentV37]):
    def _add_new_addr(self, address_mask: str, vrf: Entity | None, family: IpFamily) -> None:
        self.ip_addresses.append(IpAddressV37(
            id=0,
            display=address_mask,
            address=address_mask,
            vrf=vrf,
            prefix=None,
            family=family,
            created=datetime.now(timezone.utc),
            last_updated=datetime.now(timezone.utc),
            tags=[],
            status=Label(value="active", label="Active"),
            assigned_object_id=self.id,
        ))


@dataclass
class NetboxDeviceV37(NetboxDevice[InterfaceV37, DeviceIpV37]):
    device_role: Entity

    def __hash__(self):
        return hash((self.id, type(self)))

    def _make_interface(self, name: str, type: InterfaceType) -> InterfaceV37:
        return InterfaceV37(
            name=name,
            device=self,
            enabled=True,
            description="",
            type=type,
            id=0,
            vrf=None,
            display=name,
            untagged_vlan=None,
            tagged_vlans=[],
            ip_addresses=[],
            connected_endpoints=[],
            mode=None,
        )



from abc import abstractmethod, ABC
from collections import defaultdict
from collections.abc import Sequence
from ipaddress import ip_interface
from typing import Any, Literal, Iterable, Iterator, Optional, cast

from annet.rpl import (
    RoutingPolicy, PrefixMatchValue, SingleCondition, MatchField, RoutingPolicyStatement, ResultType,
    SingleAction, ConditionOperator, ThenField, ActionType,
)
from annet.rpl.statement_builder import NextHopActionValue, AsPathActionValue, CommunityActionValue
from .aspath import get_used_as_path_filters
from .community import get_used_united_community_lists
from .entities import (
    AsPathFilter, IpPrefixList, CommunityList, CommunityLogic, CommunityType,
    mangle_united_community_list_name, PrefixListNameGenerator, group_community_members,
)


FRR_RESULT_MAP = {
    ResultType.ALLOW: "permit",
    ResultType.DENY: "deny",
    ResultType.NEXT: "permit",
}
FRR_MATCH_COMMAND_MAP: dict[str, str] = {
    MatchField.as_path_filter: "as-path {option_value}",
    MatchField.metric: "metric {option_value}",
    MatchField.protocol: "source-protocol {option_value}",
    MatchField.interface: "interface {option_value}",
    # unsupported: as_path_length
    # unsupported: rd
}
FRR_THEN_COMMAND_MAP: dict[str, str] = {
    ThenField.local_pref: "local-preference {option_value}",
    ThenField.metric_type: "metric-type {option_value}",
    ThenField.origin: "origin {option_value}",
    ThenField.tag: "tag {option_value}",
    # unsupported: resolution
    # unsupported: rpki_valid_state
    # unsupported: mpls-label
}

FRR_INDENT = " "


class CumulusPolicyGenerator(ABC):
    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_prefix_lists(self, device: Any) -> Sequence[IpPrefixList]:
        raise NotImplementedError()

    @abstractmethod
    def get_community_lists(self, device: Any) -> list[CommunityList]:
        raise NotImplementedError()

    @abstractmethod
    def get_as_path_filters(self, device: Any) -> Sequence[AsPathFilter]:
        raise NotImplementedError

    def generate_cumulus_rpl(self, device: Any) -> Iterator[Sequence[str]]:
        prefix_lists = self.get_prefix_lists(device)
        policies = self.get_policies(device)
        prefix_list_name_generator = PrefixListNameGenerator(prefix_lists, policies)

        communities = {c.name: c for c in self.get_community_lists(device)}
        yield from self._cumulus_as_path_filters(device, policies)
        yield from self._cumulus_communities(device, communities, policies)
        yield from self._cumulus_prefix_lists(device, policies, prefix_list_name_generator)
        yield from self._cumulus_policy_config(device, communities, policies, prefix_list_name_generator)

    def _cumulus_as_path_filters(
            self,
            device: Any,
            policies: list[RoutingPolicy],
    ):
        as_path_filters = get_used_as_path_filters(self.get_as_path_filters(device), policies)
        if not as_path_filters:
            return
        for as_path_filter in as_path_filters:
            values = "_".join(x for x in as_path_filter.filters if x != ".*")
            yield "ip as-path access-list", as_path_filter.name, "permit", f"_{values}_"

    def _cumulus_prefix_list(
            self,
            ip_type: Literal["ipv6", "ip"],
            plist: IpPrefixList,
    ) -> Iterable[Sequence[str]]:
        for i, m in enumerate(plist.members):
            ge, le = m.or_longer
            yield (
                ip_type,
                "prefix-list",
                plist.name,
                f"seq {i * 5 + 5}",
                "permit", str(m.prefix),
            ) + (
                ("ge", str(ge)) if ge is not None else ()
            ) + (
                ("le", str(le)) if le is not None else ()
            )

    def _cumulus_prefix_lists(
            self, device: Any,
            policies: list[RoutingPolicy],
            name_generator: PrefixListNameGenerator,
    ) -> Iterable[Sequence[str]]:
        processed_names = set()
        for policy in policies:
            for statement in policy.statements:
                cond: SingleCondition[PrefixMatchValue]
                for cond in statement.match.find_all(MatchField.ip_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._cumulus_prefix_list("ip", plist)
                        processed_names.add(plist.name)
                for cond in statement.match.find_all(MatchField.ipv6_prefix):
                    for name in cond.value.names:
                        plist = name_generator.get_prefix(name, cond.value)
                        if plist.name in processed_names:
                            continue
                        yield from self._cumulus_prefix_list("ipv6", plist)
                        processed_names.add(plist.name)
        yield "!"

    def get_used_united_community_lists(
            self, communities: dict[str, CommunityList], policies: list[RoutingPolicy],
    ) -> list[list[CommunityList]]:
        return get_used_united_community_lists(communities=communities.values(), policies=policies)

    def _cumulus_community(
            self, name: str, cmd: str, member: str, use_regex: bool, seq: int,
    ) -> Iterable[Sequence[str]]:
        if use_regex:
            yield (
                cmd,
                "expanded",
                name,
                f"seq {seq}",
                "permit",
                member,
            )
        else:
            yield (
                cmd,
                "standard",
                name,
                f"seq {seq}",
                "permit",
                member,
            )

    def _cumulus_communities(
            self,
            device: Any,
            communities: dict[str, CommunityList],
            policies: list[RoutingPolicy],
    ) -> Iterable[Sequence[str]]:
        """ BGP community-lists section configuration """
        community_unions = self.get_used_united_community_lists(communities, policies)
        if not community_unions:
            return
        for community_list_union in community_unions:
            name = mangle_united_community_list_name([c.name for c in community_list_union])
            comm_number = 0

            for clist in community_list_union:
                if clist.type is CommunityType.BASIC:
                    member_prefix = ""
                    cmd = "bgp community-list"
                elif clist.type is CommunityType.RT:
                    member_prefix = "rt "
                    cmd = "bgp extcommunity"
                elif clist.type is CommunityType.SOO:
                    member_prefix = "soo "
                    cmd = "bgp extcommunity"
                elif clist.type is CommunityType.LARGE:
                    member_prefix = ""
                    cmd = "bgp large-community-list"
                else:
                    raise NotImplementedError(f"Community type {clist.type} is not supported on Cumulus")

                if clist.logic == CommunityLogic.AND:
                    if clist.use_regex:
                        if len(clist.members) > 1:
                            raise NotImplementedError("Multiple regexes with AND logic are not supported on Cumulus")
                        member = member_prefix + clist.members[0]
                    else:
                        member = " ".join(f"{member_prefix}{m}" for m in clist.members)
                    yield from self._cumulus_community(
                        name=name, cmd=cmd, member=member, use_regex=clist.use_regex, seq=(comm_number + 1) * 10,
                    )
                else:
                    for comm_number, member_value in enumerate(clist.members, start=comm_number):
                        yield from self._cumulus_community(
                            name=name, cmd=cmd, member=member_prefix + member_value, use_regex=clist.use_regex,
                            seq=(comm_number + 1) * 10,
                        )
                comm_number += 1
        yield "!"

    def _get_match_community_names(self, condition: SingleCondition[Sequence[str]]) -> Sequence[str]:
        if condition.operator is ConditionOperator.HAS_ANY:
            return [mangle_united_community_list_name(condition.value)]
        else:
            return condition.value

    def _cumulus_policy_match(
            self,
            device: Any,
            condition: SingleCondition[Any],
            name_generator: PrefixListNameGenerator,
    ) -> Iterator[Sequence[str]]:
        if condition.field == MatchField.community:
            for comm_name in self._get_match_community_names(condition):
                yield "match community", comm_name
            return
        if condition.field == MatchField.large_community:
            for comm_name in self._get_match_community_names(condition):
                yield "match large-community-list", comm_name
            return
        if condition.field == MatchField.extcommunity_rt:
            for comm_name in self._get_match_community_names(condition):
                yield "match extcommunity", comm_name
            return
        if condition.field == MatchField.extcommunity_soo:
            for comm_name in self._get_match_community_names(condition):
                yield "match extcommunity", comm_name
            return
        if condition.field == MatchField.ip_prefix:
            for name in condition.value.names:
                plist = name_generator.get_prefix(name, condition.value)
                yield "match", "ip address prefix-list", plist.name
            return
        if condition.field == MatchField.ipv6_prefix:
            for name in condition.value.names:
                plist = name_generator.get_prefix(name, condition.value)
                yield "match", "ipv6 address prefix-list", plist.name
            return
        if condition.operator is not ConditionOperator.EQ:
            raise NotImplementedError(
                f"`{condition.field}` with operator {condition.operator} is not supported for Cumulus",
            )
        if condition.field not in FRR_MATCH_COMMAND_MAP:
            raise NotImplementedError(f"Match using `{condition.field}` is not supported for Cumulus")
        cmd = FRR_MATCH_COMMAND_MAP[condition.field]
        yield "match", cmd.format(option_value=condition.value)

    def _cumulus_then_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set community together with add/replace on cumulus",
                )
            members = [m for name in action.value.replaced for m in communities[name].members]
            if members:
                yield "set", "community", *members
            else:
                yield "set", "community", "none"
        if action.value.added:
            members = [m for name in action.value.added for m in communities[name].members]
            yield "set", "community", *members, "additive"
        for community_name in action.value.removed:
            yield "set comm-list", community_name, "delete"

    def _cumulus_then_large_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            raise NotImplementedError("Replacing Large community is not supported for Cumulus")
        for community_name in action.value.added:
            yield "set", "large-community", community_name, "additive"
        for community_name in action.value.removed:
            raise NotImplementedError("Large-community remove is not supported for Cumulus")

    def _cumulus_then_rt_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            raise NotImplementedError("Replacing RT extcommunity is not supported for Cumulus")
        for community_name in action.value.added:
            yield "set", "extcommunity rt", community_name, "additive"
        for community_name in action.value.removed:
            raise NotImplementedError("RT extcommunity remove is not supported for Cumulus")

    def _cumulus_then_soo_community(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.replaced is not None:
            raise NotImplementedError("Replacing SOO extcommunity is not supported for Cumulus")
        for community_name in action.value.added:
            yield "set", "extcommunity soo", community_name, "additive"
        if action.value.removed:
            raise NotImplementedError("SOO extcommunity remove is not supported for Cumulus")

    def _cumulus_extcommunity_type_str(self, comm_type: CommunityType) -> str:
        if comm_type is CommunityType.SOO:
            return "soo"
        elif comm_type is CommunityType.RT:
            return "rt"
        elif comm_type is CommunityType.LARGE:
            raise ValueError("Large community is not subtype of extcommunity")
        elif comm_type is CommunityType.BASIC:
            raise ValueError("Basic community is not subtype of extcommunity")
        else:
            raise NotImplementedError(f"Community type {comm_type} is not supported on cumulus")

    def _cumulus_then_extcommunity(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[CommunityActionValue],
    ):
        if action.value.replaced is not None:
            if action.value.added or action.value.removed:
                raise NotImplementedError(
                    "Cannot set extcommunity together with add/delete on cumulus",
                )
            if not action.value.replaced:
                yield "set", "extcommunity", "none"
                return
            members = group_community_members(communities, action.value.replaced)
            for community_type, replaced_members in members.items():
                type_str = self._cumulus_extcommunity_type_str(community_type)
                yield "set", "extcommunity", type_str, *replaced_members
        if action.value.added:
            raise NotImplementedError("extcommunity add is not supported for Cumulus")
        if action.value.removed:
            raise NotImplementedError("extcommunity remove is not supported for Cumulus")

    def _cumulus_then_as_path(
            self,
            device: Any,
            action: SingleAction[AsPathActionValue],
    ) -> Iterator[Sequence[str]]:
        if action.value.prepend:
            for path_item in action.value.prepend:
                yield "set", "as-path prepend", path_item
        if action.value.expand:
            raise NotImplementedError("asp_path.expand is not supported for Cumulus")
        if action.value.delete:
            for path_item in action.value.delete:
                yield "set", "as-path exclude", path_item
        if action.value.set is not None:
            yield "set", "as-path exclude", "all"
            for path_item in action.value.set:
                yield "set", "as-path prepend", path_item
        if action.value.expand_last_as:
            yield "set", "as-path prepend last-as", action.value.expand_last_as

    def _cumulus_policy_then(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            action: SingleAction[Any],
    ) -> Iterator[Sequence[str]]:
        if action.field == ThenField.community:
            yield from self._cumulus_then_community(
                communities,
                device,
                cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.large_community:
            yield from self._cumulus_then_large_community(
                communities,
                device,
                cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.extcommunity:
            yield from self._cumulus_then_extcommunity(communities, device, action)
            return
        if action.field == ThenField.extcommunity_rt:
            yield from self._cumulus_then_rt_community(
                communities,
                device,
                cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.extcommunity_soo:
            yield from self._cumulus_then_soo_community(
                communities,
                device,
                cast(SingleAction[CommunityActionValue], action),
            )
            return
        if action.field == ThenField.metric:
            if action.type is ActionType.ADD:
                yield "set", f"metric +{action.value}"
            elif action.type is ActionType.REMOVE:
                yield "set", f"metric -{action.value}"
            elif action.type is ActionType.SET:
                yield "set", f"metric {action.value}"
            else:
                raise NotImplementedError(f"Action type {action.type} for metric is not supported for Cumulus")
            return
        if action.field == ThenField.as_path:
            yield from self._cumulus_then_as_path(device, action)
            return
        if action.field == ThenField.next_hop:
            next_hop_action_value = cast(NextHopActionValue, action.value)
            if next_hop_action_value.target == "self":
                yield "set", "metric 1"
            elif next_hop_action_value.target == "discard":
                pass
            elif next_hop_action_value.target == "peer":
                pass
            elif next_hop_action_value.target == "ipv4_addr":
                yield "set", f"ip next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "ipv6_addr":
                yield "set", f"ipv6 next-hop {next_hop_action_value.addr}"
            elif next_hop_action_value.target == "mapped_ipv4":
                yield "set", "ipv6 next-hop ::FFFF:{next_hop_action_value.addr}"
            else:
                raise NotImplementedError(
                    f"Next_hop target {next_hop_action_value.target} is not supported for Cumulus")
            return

        if action.type is not ActionType.SET:
            raise NotImplementedError(f"Action type {action.type} for `{action.field}` is not supported for Cumulus")
        if action.field not in FRR_THEN_COMMAND_MAP:
            raise NotImplementedError(f"Then action using `{action.field}` is not supported for Cumulus")
        cmd = FRR_THEN_COMMAND_MAP[action.field]
        yield "set", cmd.format(option_value=action.value)

    def _cumulus_policy_statement(
            self,
            communities: dict[str, CommunityList],
            device: Any,
            policy: RoutingPolicy,
            statement: RoutingPolicyStatement,
            prefix_list_name_generator: PrefixListNameGenerator,
    ) -> Iterable[Sequence[str]]:
        yield "route-map", policy.name, FRR_RESULT_MAP[statement.result], str(statement.number)

        for condition in statement.match:
            for row in self._cumulus_policy_match(device, condition, prefix_list_name_generator):
                yield FRR_INDENT, *row
        for action in statement.then:
            for row in self._cumulus_policy_then(communities, device, action):
                yield FRR_INDENT, *row
        if statement.result is ResultType.NEXT:
            yield FRR_INDENT, "on-match next"
        yield "!"

    def _cumulus_policy_config(
            self,
            device: Any,
            communities: dict[str, CommunityList],
            policies: list[RoutingPolicy],
            prefix_list_name_generator: PrefixListNameGenerator,
    ) -> Iterable[Sequence[str]]:
        """ Route maps configuration """

        for policy in policies:
            applied_stmts: dict[int, Optional[str]] = {}
            for statement in policy.statements:
                if statement.number is None:
                    raise RuntimeError(
                        f"Statement number should not be empty on Cumulus (found for policy: {policy.name})"
                    )

                if statement.number in applied_stmts:
                    raise RuntimeError(
                        f"Multiple statements have same number {statement.number} for policy `{policy.name}`: "
                        f"`{statement.name}` and `{applied_stmts[statement.number]}`")
                yield from self._cumulus_policy_statement(
                    communities, device, policy, statement, prefix_list_name_generator,
                )
                applied_stmts[statement.number] = statement.name



import abc
import dataclasses
import enum
import io
import json
import os
import sys
import types
from collections import OrderedDict as odict
from contextlib import contextmanager
from typing import Any, List, Pattern

import colorama
import pygments
import pygments.formatters
import pygments.lexers.data
import yaml

LABEL_NEW_PREFIX = "new: "


# =====
class OutputWriter:
    def __init__(self, data):
        self.data = data

    def write(self, file):
        if isinstance(self.data, types.GeneratorType):
            for n in self.data:
                file.write(n)
            return
        file.write(self.data)

        if self.data and self.data[-1:] != "\n":
            file.write("\n")


class YamlWriter(OutputWriter):
    def write(self, file):
        print_as_yaml(self.data, file)


class JsonWriter(OutputWriter):
    def write(self, file):
        print_as_json(self.data, file)


def dir_or_file_output(dest, items_count, create_dir=True, suggest_dir=False):
    if dest is None:
        dest = ""

    dir_mode = dest.endswith(os.sep) or suggest_dir or dest not in ("-", "") and items_count > 1

    if os.path.exists(dest):
        is_dir = os.path.isdir(dest)
        if dir_mode and not is_dir:
            raise EnvironmentError(
                f"Output set to dir {dest}, but file with the same name is on the way"
            )
        dir_mode = dir_mode or is_dir
    elif create_dir and dest not in ("-", ""):
        pdir = dest if dir_mode else os.path.dirname(dest)
        if pdir:
            os.makedirs(pdir, exist_ok=True)

    return dir_mode


def print_label(label, back_color=colorama.Back.GREEN, file=None):
    if file is None:
        file = sys.stdout
    tty = bool(os.isatty(file.fileno()))
    print("%s%s# %s %s %s%s%s" % (
        colorama.Style.BRIGHT * tty, back_color * tty,
        "-" * 20, label, "-" * 20,
        colorama.Back.RESET * tty, colorama.Style.RESET_ALL * tty,
    ), file=file)


def print_err_label(label):
    print_label(label, colorama.Back.RED)


def print_as_yaml(data, fh=sys.stdout):
    class UnsortableTuple(tuple):
        def __lt__(self, _):
            raise TypeError

    class Dumper(yaml.Dumper):
        def ignore_aliases(self, data):
            return True

        def represent_ordered_dict(self, data):
            return self.represent_dict([UnsortableTuple(item) for item in data.items()])

    Dumper.add_representer(odict, Dumper.represent_ordered_dict)

    def yaml_dump(data, **kwargs):
        return yaml.dump(data, allow_unicode=True, indent=4, Dumper=Dumper, **kwargs)

    if fh.isatty():
        pygments.highlight(
            code=yaml_dump(data, width=os.get_terminal_size().columns),
            lexer=pygments.lexers.data.YamlLexer(),
            formatter=pygments.formatters.TerminalFormatter(bg="dark"),
            outfile=fh,
        )
    else:
        fh.write(yaml_dump(data))


class Dumpable(abc.ABC):
    def dump(self) -> Any:
        pass


def print_as_json(data, fh=sys.stdout):
    """
    В выводе dict ключи отсортированы по имени, кроме ключей в OrderedDict,
    которые сохраняют свой оригинальный порядок
    """

    class ODictKey(str):
        """
        Строка с дополнительным параметром :index, по которому сортируется
        """

        @staticmethod
        def __new__(cls, str_value, index=0):
            self = str.__new__(cls, str_value)
            self.ind = index
            return self

        def __lt__(self, other):
            assert type(self) is type(other)
            return (self.ind, self) < (other.ind, other)  # pylint: disable=no-member

    class UnsortableOdict(odict):
        """
        Разновидность OrderedDict, ключи которого сохраняют порядок
        при попытке строковой сортировки (содержат доп. индекс)
        """

        def __init__(self, orig):
            super().__init__((ODictKey(key, index), value)
                             for (index, (key, value)) in enumerate(orig.items()))

        @staticmethod
        def convert_odicts(data):
            self = UnsortableOdict.convert_odicts
            if isinstance(data, dict):
                data = type(data)((key, self(value)) for (key, value) in data.items())
            elif isinstance(data, (tuple, list)):
                data = type(data)(self(value) for value in data)
            if isinstance(data, odict):
                data = UnsortableOdict(data)
            return data

    class JsonEncoder(json.JSONEncoder):
        def default(self, o):
            # https://github.com/PyCQA/pylint/issues/3537
            if isinstance(o, Pattern):  # pylint: disable=isinstance-second-argument-not-valid-type
                return "(regex obj)/" + o.pattern + "/ (not a string)"
            elif isinstance(o, Dumpable):
                return o.dump()
            elif dataclasses.is_dataclass(o):
                return dataclasses.asdict(o)
            elif isinstance(o, enum.Enum):
                return o.value

            raise TypeError("can't serialize %r to json" % o)

        def encode(self, o):
            try:
                return super().encode(o)
            except TypeError:
                raise TypeError("can't serialize %r to json" % o)

    dump = json.dumps(UnsortableOdict.convert_odicts(data), cls=JsonEncoder,
                      ensure_ascii=False, indent=4, sort_keys=True) + "\n"
    if fh.isatty():
        pygments.highlight(
            code=dump,
            lexer=pygments.lexers.data.JsonLexer(),
            formatter=pygments.formatters.TerminalFormatter(bg="dark"),
            outfile=fh,
        )
    else:
        fh.write(dump)


class TextArgs:
    __slots__ = ("text", "color", "offset")

    def __init__(self, text, color=None, offset=None):
        self.text = text
        self.color = color
        self.offset = offset  # смещение от начала линии

    def __repr__(self):
        return "%s(%r, %s, %s)" % (self.__class__.__name__, self.text, self.color, self.offset)


# =====
@contextmanager
def capture_output(new_stdout: io.BufferedRandom = None, new_stderr: io.BufferedRandom = None):
    streams = (
        (sys.stdout, new_stdout),
        (sys.stderr, new_stderr),
    )
    backup = []
    for (std_stream, new_stream) in streams:
        if new_stream is not None:
            new_stream.seek(0)
            new_stream.truncate(0)
            std_fd = std_stream.fileno()
            std_stream.flush()
            backup.append((std_fd, os.fdopen(os.dup(std_fd), "w"), new_stream))
            os.dup2(new_stream.fileno(), std_fd)
    try:
        yield
    finally:
        for (std_fd, backup_stream, stream) in backup:
            backup_stream.flush()
            os.dup2(backup_stream.fileno(), std_fd)
            backup_stream.close()
            stream.seek(0)


def format_file_diff(diff_lines: List[str]) -> List[str]:
    colors = (
        ("+++", colorama.Fore.CYAN),
        ("---", colorama.Fore.CYAN),
        ("@@", colorama.Fore.CYAN),
        ("+", colorama.Fore.GREEN),
        ("-", colorama.Fore.RED),
    )
    ret = []
    for line in diff_lines:
        for (check, color) in colors:
            if line.startswith(check):
                ret.append(color + line + colorama.Fore.RESET)
                break
        else:
            ret.append(line)
    return ret
